"""Utilities for dealing with package version info.

See also stsci.distutils.svnutils which specifically deals with adding SVN
info to version.py modules.
"""




import datetime
import os
import subprocess

from .astutils import ImportVisitor, walk


VERSION_PY_TEMPLATE = """
\"\"\"This is an automatically generated file created by %(hook_function)s.
Do not modify this file by hand.
\"\"\"

__all__ = ['__version__', '__vdate__', '__svn_revision__', '__svn_full_info__',
           '__setup_datetime__']

import datetime

__version__ = %(version)r
__vdate__ = %(vdate)r
__svn_revision__ = %(svn_revision)r
__svn_full_info__ = %(svn_full_info)r
__setup_datetime__ = %(setup_datetime)r

# what version of stsci.distutils created this version.py
stsci_distutils_version = %(stsci_distutils_version)r

if '.dev' in __version__:
    def update_svn_info():
        \"\"\"Update the SVN info if running out of an SVN working copy.\"\"\"

        import os
        import string
        import subprocess

        global __svn_revision__
        global __svn_full_info__

        path = os.path.abspath(os.path.dirname(__file__))

        run_svnversion = True

        try:
            pipe = subprocess.Popen(['svn', 'info', path],
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
            stdout, _ = pipe.communicate()
            if pipe.returncode == 0:
                lines = []
                for line in stdout.splitlines():
                    line = line.decode('latin1').strip()
                    if not line:
                        continue
                    lines.append(line)

                if not lines:
                    __svn_full_info__ = ['unknown']
                else:
                    __svn_full_info__ = lines
            else:
                run_svnversion = False
        except OSError:
            run_svnversion = False

        if run_svnversion:
            # If updating the __svn_full_info__ succeeded then use its output
            # to find the base of the working copy and use svnversion to get
            # the svn revision.
            for line in __svn_full_info__:
                if line.startswith('Working Copy Root Path'):
                    path = line.split(':', 1)[1].strip()
                    break

            try:
                pipe = subprocess.Popen(['svnversion', path],
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE)
                stdout, _ = pipe.communicate()
                if pipe.returncode == 0:
                    stdout = stdout.decode('latin1').strip()
                    if stdout and stdout[0] in string.digits:
                        __svn_revision__ = stdout
            except OSError:
                pass

        # Convert __svn_full_info__ back to a string
        if isinstance(__svn_full_info__, list):
            __svn_full_info__ = '\\n'.join(__svn_full_info__)


    update_svn_info()
    del update_svn_info
"""[1:]


def package_uses_version_py(package_root, package, module_name='version'):
    """Determines whether or not a version.py module should exist in the given
    package.  Returns the full path to the version.py module, regardless of
    whether it already exists.  Otherwise returns False.

    This works by checking whether or not there are any imports from the
    'version' module in the package's ``__init__.py``.

    You should write this in your package's ``__init__.py``::

        from .version import *

    """

    pdir = os.path.join(package_root, *(package.split('.')))
    init = os.path.join(pdir, '__init__.py')
    if not os.path.exists(init):
        raise Exception('Not a valid package - no __init__.py')

    try:
        visitor = ImportVisitor()
        walk(init, visitor)
    except:
        raise SyntaxError('Not able to parse %s' % init)

    found = False
    # Check the import statements parsed from the file for an import of or
    # from the svninfo module in this package
    for imp in visitor.imports:
        if imp[0] in (module_name, '.'.join((package, module_name))):
            found = True
            break
    for imp in visitor.importfroms:
        mod = imp[0]
        name = imp[1]
        if (mod in (module_name, '.'.join((package, module_name))) or
            (mod == package and name == module_name)):
            found = True
            break

    if not found:
        return False

    return os.path.join(pdir, module_name + '.py')


def clean_version_py(package_dir, package):
    """Removes the generated version.py module from a package, but only if
    we're in an SVN working copy.
    """

    pdir = os.path.join(package_root, *(package.split('.')))
    version_py = os.path.join(pdir, 'version.py')
    if not os.path.exists(svninfo):
        return

    try:
        pipe = subprocess.Popen(['svn', 'status', svninfo],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
    except OSError:
        return

    if pipe.wait() != 0:
        return

    # TODO: Maybe don't assume ASCII here.  Find out the best way to handle
    # this.
    if not pipe.stdout.read().decode('latin1').startswith('?'):
        return

    os.remove(version_py)


def update_setup_datetime(filename='version.py'):
    """Update the version.py with the last time a setup command was run."""

    if not os.path.exists(filename):
        return

    d = datetime.datetime.now()

    lines = []
    with open(filename, 'r') as f:
        lines = f.readlines()

    with open(filename, 'w') as f:
        for line in lines:
            if not line.startswith('__setup_datetime__'):
                f.write(line)
            else:
                f.write('__setup_datetime__ = %r\n' % d)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, absolute_import

import os

from astropy.io import fits

import numpy.ma as ma
import numpy as np

from . import exposure as exp
from .psf_library import PSFLibrary
from .telescope import Telescope
from .instrument import Instrument
from .io_utils import ref_data_interp, read_json, get_json_parameters
from .custom_exceptions import EngineInputError, RangeError, DataError


class JWST(Telescope):

    """
    This is currently a dummy class that is used for configuration file discovery. Could eventually
    contain JWST-specific methods.
    """
    pass


class JWSTInstrument(Instrument):

    """
    Generic JWST Instrument class
    """

    def __init__(self, mode=None, config={}, webapp=False, **kwargs):
        telescope = JWST()
        # these are the required sections and need to be passed via API in webapp mode
        self.instrument_pars = {}
        self.instrument_pars['detector'] = ["nexp", "ngroup", "nint", "readmode", "subarray"]
        self.instrument_pars['instrument'] = ["aperture", "disperser", "filter", "instrument", "mode"]
        self.api_parameters = list(self.instrument_pars.keys())

        # these are required for calculation, but ok to live with config file defaults
        if hasattr(self, 'api_ignore'):
            self.api_ignore.extend(['dynamic_scene', 'max_scene_size', 'scene_size'])
        else:
            self.api_ignore = ['dynamic_scene', 'max_scene_size', 'scene_size']

        Instrument.__init__(self, telescope=telescope, mode=mode, config=config, webapp=webapp, **kwargs)


class NIRSpec(JWSTInstrument):

    """
    Need to over-ride get_wave_range() for NIRSpec because the effective wavelength range
    depends on the blocking filter, the aperture, and the disperser.  Also need to overload __init__
    to handle some special MSA configuration needs.
    """

    def __init__(self, mode=None, config={}, webapp=False, **kwargs):

        # Needed for 'rn' json detector parameters
        self.detector_readmode = config['detector']['readmode']

        JWSTInstrument.__init__(self, mode=mode, config=config, webapp=webapp, **kwargs)

        if self.mode == "msa":
            aperture = self.instrument['aperture']
            shutter_location = self.instrument['shutter_location']
            gap_config_file = self.aperture_config[aperture].pop('gap')
            gap_config = read_json(os.path.join(self.ref_dir, gap_config_file), raise_except=True)
            self.shutter_locations = gap_config.keys()
            try:
                self.aperture_config[aperture]["gap"] = gap_config[shutter_location]
            except KeyError as e:
                msg = "Shutter location not specified for MSA calculation: %s" % e
                raise DataError(value=msg)

    def get_wave_range(self):
        """
        Get the wavelength range of the current instrument configuration

        Returns
        -------
        range_dict: dict
            Contains the instrument wavelength range in microns described by:
                wmin - minimum wavelength
                wmax - maximum wavelength
        """
        disperser = self.instrument['disperser']
        aperture = self.instrument['aperture']
        filt = self.instrument['filter']
        gap = self.aperture_config[aperture]['gap']

        if disperser is not None:
            # get the wavelength range from the gap configuration
            if filt in gap[disperser]:
                # g140m and g140h have different gap configs for each blocking filter
                g_wmin = gap[disperser][filt]["wave_min"]
                g_wmax = gap[disperser][filt]["wave_max"]
            else:
                g_wmin = gap[disperser]["wave_min"]
                g_wmax = gap[disperser]["wave_max"]

            # get the wavelength range from the configuration file
            c_wmin = self.range[aperture][filt]["wmin"]
            c_wmax = self.range[aperture][filt]["wmax"]

            # get the wavelength range over which the disperser efficiency is known
            wave_blaze = self.get_wave_blaze()
            d_wmin = wave_blaze.min()
            d_wmax = wave_blaze.max()

            # get the wavelength range over which the filter throughput is known
            wave_filter = self.get_wave_filter()
            f_wmin = wave_filter.min()
            f_wmax = wave_filter.max()

            # compare filter and disperser wavelength ranges
            if f_wmax < d_wmin or d_wmax < f_wmin:
                raise RangeError(value="Disperser and Filter wavelength ranges do not overlap.")
            wmin = max(f_wmin, d_wmin, c_wmin, g_wmin)
            wmax = min(f_wmax, d_wmax, c_wmax, g_wmax)
        else:
            wmin = self.range[aperture][filt]["wmin"]
            wmax = self.range[aperture][filt]["wmax"]

        range_dict = {'wmin': wmin, 'wmax': wmax}
        return range_dict

    def create_gap_mask(self, wave):
        """
        Use the gap configuration and a wavelength vector, wave, to build a masked array that
        masks out the location of the gap.  Wavelengths between and including both gap endpoints
        will be masked out.

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to construct mask from

        Returns
        -------
        mask: numpy.ma 1D masked array
            1D array masked at the locations within the configured detector gap and 1.0 elsewhere
        """
        disperser = self.instrument['disperser']
        aperture = self.instrument['aperture']
        filt = self.instrument['filter']
        gap = self.aperture_config[aperture]['gap'][disperser]
        if filt in gap:
            gap_start = gap[filt]['gap_start']
            gap_end = gap[filt]['gap_end']
        else:
            gap_start = gap['gap_start']
            gap_end = gap['gap_end']

        if gap_start is not None and gap_end is not None:
            masked_wave = ma.masked_inside(wave, gap_start, gap_end)
            mask = masked_wave / wave
            mask = ma.filled(mask, 0.0)
        else:
            mask = 1.0
        return mask

    def get_internal_eff(self, wave):
        """
        Read in internal efficiency of NIRSpec. This is
        overloaded because the internal optical throughput is
        different for the NIRSpec IFU compared to MOS and Fixed Slit.

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto

        Returns
        -------
        eff: numpy.ndarray or float
            Internal throughput as a function of wave
        """
        if self.mode == "ifu":
            eff = self._get_throughput(wave, 'internal_ifu')
        elif self.mode in ["msa", "fixed_slit", "target_acq"]:
            eff = self._get_throughput(wave, 'internal_msa')
        else:
            msg = "Internal efficiency not configured for NIRSpec mode %s." % self.mode
            raise EngineInputError(value=msg)

        return eff

    def get_detector_pars(self):
        """
        Copied this here but may not need to override.
        """

        det_dict = {key: get_json_parameters(self.det_pars, key, readmode=self.detector_readmode)
                    for key in self.det_pars.keys() if not key == 'meta'}

        return det_dict


class NIRCam(JWSTInstrument):

    def _loadpsfs(self):
        """
        For the bar-shaped coronagraphy masks we need to load the psf_library on a per-filter basis.
        The PSF files have the aperture and filter smooshed into one string so use that as the key.
        """
        if self.instrument['aperture'] in ('masklwb', 'maskswb'):
            psfs_key = "%s%s" % (self.instrument['aperture'], self.instrument['filter'])
        else:
            psfs_key = self.instrument['aperture']

        psf_path = os.path.join(self.ref_dir, "psfs")
        self.psf_library = PSFLibrary(path=psf_path, aperture=psfs_key)

    def get_filter_eff(self, wave):
        """
        over-ride the filter efficiency because the narrow-band filters are in the pupil wheel,
        and therefore also go through a broad-band filter in the filter wheel (which doesn't have a clear).

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto

        Returns
        -------
        eff: numpy.ndarray or float
            Filter throughput as a function of wave
        """

        if not hasattr(self, 'double_filters'):
            msg = "NIRCam requires a mapping that describes which filters are actually a combination of two filters."
            raise DataError(value=msg)

        if self.instrument['filter'] in self.double_filters.keys():
            eff = self._get_throughput(wave, self.instrument['filter'])
            eff_pupil = self._get_throughput(wave, self.double_filters[self.instrument['filter']])
            eff *= eff_pupil
        else:
            eff = self._get_throughput(wave, self.instrument['filter'])
        return eff

    def get_detector_pars(self):
        """
        Need to over-ride get_detector_pars() to handle the two different detectors. Which one to use is keyed off
        of the configured aperture.
        """
        aperture = self.instrument['aperture']
        if aperture in ['lw', 'mask335r', 'mask430r', 'masklwb']:
            aperture_lu = 'lw'
        elif aperture in ['sw', 'mask210r', 'maskswb']:
            aperture_lu = 'sw'
        else:
            msg = "Unknown NIRCam aperture %s" % aperture
            raise EngineInputError(value=msg)

        det_dict = {key: get_json_parameters(self.det_pars, key, aperture=aperture_lu)
                    for key in self.det_pars.keys() if not key == 'meta'}

        return det_dict

    def get_internal_eff(self, wave):
        """
        Read in internal efficiency. For NIRCam there are separate internal efficiencies for the optics common
        to all modes, the throughput of the coronagrapher substrate, and the throughputs of the optical wedges
        that bring the coronagraphy elements into the field of view of the detectors.

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto

        Returns
        -------
        eff: numpy.ndarray or float
            Internal throughput as a function of wave
        """
        base_eff = self._get_throughput(wave, 'internal')
        coronagraphy_eff = 1.0
        wedge_eff = 1.0
        dichroic_eff = 1.0

        # load the dichroic throughput based on which channel we're using
        ap = self.aperture_config[self.instrument['aperture']]
        if 'channel' not in ap:
            msg = "Channels not configured for NIRCam aperture %s." % self.instrument['aperture']
            raise DataError(value=msg)

        if ap['channel'] == 'sw':
            dichroic_eff = self._get_throughput(wave, 'dbs_sw')
        elif ap['channel'] == 'lw':
            dichroic_eff = self._get_throughput(wave, 'dbs_lw')
        else:
            msg = "Channel %s not configured for NIRCam aperture %s." % (ap['channel'], self.instrument['aperture'])
            raise DataError(value=msg)

        if self.instrument['mode'] == 'coronagraphy':
            coronagraphy_eff = self._get_throughput(wave, 'coronagraphy_substrate')
            if self.instrument['aperture'] in ('maskswb', 'mask210r'):
                wedge_eff = self._get_throughput(wave, 'sw_wedge_eff')
            if self.instrument['aperture'] == ('masklwb', 'mask335r', 'mask430r'):
                wedge_eff = self._get_throughput(wave, 'lw_wedge_eff')
        eff = base_eff * coronagraphy_eff * wedge_eff * dichroic_eff

        return eff

    def dispersion_axis(self):
        """
        The dispersion axis is either along rows (the X axis) or along columns (Y axis).  By default
        it is along the X axis.  However, the GRISMC grating disperses along the Y axis as a way to
        help mitigate source crowding and confusion.

        Returns
        -------
        disp_axis: str
            Axis along with spectra are dispersed.  Allowed values are 'x' or 'y'.
        """
        if self.projection_type == 'slitless' and self.instrument['disperser'] == "grismc":
            disp_axis = "y"
        else:
            disp_axis = "x"
        return disp_axis

    def bar_width(self, x):
        """
        Width of MASKLWB or MASKSWB in arcsec as a function of X. The width at the center of the FOV is taken from the
        configuration as a function of what filter is being used.

        Parameters
        ----------
        x: float
            X position (arcsec) in field of view

        Returns
        ------
        width: float
            Width of bar at X in arcsec
        """
        filt = self.instrument['filter']
        if filt not in self.bar_offsets:
            msg = "Invalid filter, %s, for MASKLWB/MASKSWB." % filt
            raise DataError(value=msg)
        # change the sign because webbpsf has X increase to left while we have it increasing to the right.
        # the bars narrow from left to right in the FOV.
        center = -1.0 * self.bar_offsets[filt]
        if self.instrument['aperture'] == 'maskswb':
            width = 0.2666 - 0.01777 * (center + x)
        elif self.instrument['aperture'] == 'masklwb':
            width = 0.5839 - 0.03893 * (center + x)
        else:
            msg = "bar_width() method only appropriate for MASKLWB and MASKSWB apertures."
            raise EngineInputError(value=msg)
        return width

    def get_detector_qe(self, wave):
        """
        Need to over-ride get_detector_qe() to handle the two different detectors. Which one to use is keyed off
        of the configured aperture.
        """
        aperture = self.instrument['aperture']
        try:
            channel = self.aperture_config[aperture]['channel']
        except KeyError as e:
            msg = "NIRCam aperture configuration must include which channel the aperture belongs to, sw or lw. (%s)" % e
            raise DataError(value=msg)
        if channel == 'lw':
            qe = self._get_throughput(wave, 'qe_lw')
        elif channel == 'sw':
            qe = self._get_throughput(wave, 'qe_sw')
        else:
            msg = "Unknown NIRCam channel %s" % channel
            raise DataError(value=msg)

        return qe


class NIRISS(JWSTInstrument):

    """
    Need to over-ride get_internal_eff() because there are two different wheels in NIRISS.
    The 'filter wheel' has a clear position with 100% transmission. The 'pupil wheel'
    (which somewhat confusingly also contains filters) has a clear position with some
    obstruction (the Pupil Alignment Reference - PAR). So if the active filter is in the
    'filter wheel', the transmission takes a hit from the PAR in the pupil wheel.

    Also need to override dispersion_axis() because the BG150C grating disperses along the
    Y axis vs. the X axis like every other JWST disperser.
    """

    def __init__(self, mode=None, config={}, webapp=False, **kwargs):

        # these are required for calculation, but ok to live with config file defaults
        self.api_ignore = ['max_saturated_pixels', 'min_snr_threshold', 'aperture_size']

        JWSTInstrument.__init__(self, mode=mode, config=config, webapp=webapp, **kwargs)

    def dispersion_axis(self):
        """
        The dispersion axis is either along rows (the X axis) or along columns (Y axis).  By default
        it is along the X axis.  However, the GR150C grating disperses along the Y axis as a way to
        help mitigate source crowding and confusion.

        Returns
        -------
        disp_axis: str
            Axis along with spectra are dispersed.  Allowed values are 'x' or 'y'.
        """
        if self.instrument['disperser'] in ("gr150c", "gr700xd"):
            disp_axis = "y"
        else:
            disp_axis = "x"
        return disp_axis

    def get_internal_eff(self, wave):
        """
        Calculate NIRISS internal efficiency

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate efficiency onto

        Returns
        -------
        eff: numpy.ndarray or float
            Internal efficiency as a function of wave
        """
        filt = self.instrument['filter']
        if self.mode in ['ami', 'imaging']:
            # Pupil wheel
            if filt in ["f090w", "f115w", "f158m", "f140m", "f150w", "f200w"]:
                pupil_file = os.path.join(self.ref_dir, self.paths["clear"])
            # Filter wheel
            elif filt in ["f277w", "f444w", "f356w", "f430m", "f380m", "f480m"]:
                pupil_file = os.path.join(self.ref_dir, self.paths["clearp"])

            pupil_eff = ref_data_interp(pupil_file, wave, 'throughput')
        else:
            pupil_eff = 1.0

        # handle NRM transmission
        if self.mode == 'ami':
            nrm_file = os.path.join(self.ref_dir, self.paths["nrm"])
            nrm_eff = ref_data_interp(nrm_file, wave, 'throughput')
        else:
            nrm_eff = 1.0

        internal_file = os.path.join(self.ref_dir, self.paths["internal"])
        internal_eff = ref_data_interp(internal_file, wave, 'throughput')

        # Total internal efficiency are the internal reflections and the pupil/filter wheel transmissions
        eff = internal_eff * pupil_eff * nrm_eff

        return eff

    def get_extraction_mask(self, order):
        """
        Each SOSS order has its own extraction mask. Use the specified order to build the key to look up the
        FITS file containing the mask.

        Parameters
        ----------
        order: int
            Order whose mask to read

        Returns
        -------
        mask: 2D np.ndarray
            Mask data
        """
        if order not in (1, 2, 3):
            msg = "SOSS order %d is not valid." % order
            raise EngineInputError(value=msg)

        key = "gr700xd_%d_mask" % order

        # substrip96 is a special case where there's only one possible mask
        if self.detector['subarray'] == 'substrip96':
            key += "96"

        path = self.paths.get(key, None)

        if path is None:
            msg = "No mask configured for SOSS order %d." % order
            raise DataError(value=msg)

        mask_file = os.path.join(self.ref_dir, path)
        try:
            mask = fits.getdata(mask_file)
        except Exception as e:
            msg = "Error reading mask data for SOSS order %d: %s" % (order, type(e))
            raise DataError(value=msg)

        return mask

    def get_trace(self, wave):
        """
        Read in spectral trace offsets from center of FOV. Currently wavelength-dependent spatial distortion is
        only required for SOSS mode. Other modes simply return 0's.

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate/trim trace data onto

        Returns
        -------
        trace: numpy.ndarray or float
            Spectral trace offset from center of FOV as a function of wave
        """
        if self.mode == "soss":
            disperser = self._get_disperser()
            key = "%s_wavepix" % disperser
            # handle the special case of substrip96
            if self.detector['subarray'] == 'substrip96':
                key += "96"

            # SOSS modes requires trace reference data to work properly. so raise exception if we can't load it.
            try:
                trace = self._interp_refdata(wave, key, colname='trace')
            except DataError as e:
                msg = "Spectral trace data missing for NIRISS SOSS, %s."
                raise DataError(value=msg)
        else:
            trace = np.zeros_like(wave)
        return trace

    def get_detector_pixels(self, wave):
        """
        Read in detector pixel positions for each wavelength in wave_pix

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate/trim pixel position data onto

        Returns
        -------
        pixels: numpy.ndarray or None
            Detector pixel positions corresponding to each element of wave
        """
        disperser = self._get_disperser()
        if disperser is not None:
            key = "%s_wavepix" % disperser
            # handle the special case of substrip96
            if self.detector['subarray'] == 'substrip96':
                key += "96"
            try:
                pixels = self._interp_refdata(wave, key, colname='detector_pixels', default=np.arange(len(wave)))
            except DataError as e:
                pixels = None
        else:
            pixels = None
        return pixels


    def get_detector_pars(self):
        """
        Copied this here but may not need to override.
        """

        det_dict = {key: get_json_parameters(self.det_pars, key) for key in self.det_pars.keys() if not key == 'meta'}

        return det_dict


class MIRI(JWSTInstrument):

    """
    The MIRI internal efficiency, detector readmodes, etc. are more complex, and different than the other instruments,
    so some methods are redefined.
    """

    @property
    def qe_key(self):
        """
        MIRI has three different detectors with three different QE reference files.  Use this
        method to pick the right reference file key based on the configured aperture and overload self.qe_key.

        Returns
        -------
        key: string
            Key for looking up the appropriate QE reference file
        """
        aperture = self.instrument["aperture"]
        if aperture in ['ch1', 'ch2']:
            key = "miri_sw_qe"
        elif aperture in ['ch3', 'ch4']:
            key = "miri_lw_qe"
        else:
            key = "miri_imager_qe"
        return key

    def get_detector_pars(self):
        """
        Get detector parameters which are keyed by short wavelength (sw), long wavelength (lw),
        and imager.  Use configured disperser and aperture to figure out which to use.

        Returns
        -------
        det_dict: dict
            rn - readout noise (float)
            ipc - detector has inter-pixel capacitance (bool)
            dark_current - dark current (float)
            fullwell - pixel capacity in electrons (float)
            rn_correlation - detector has correlated readout noise (bool)
            ff_electrons - number of counts in flat-field frame (float)
        """

        disperser = self.instrument["disperser"]
        aperture = self.instrument["aperture"]

        if aperture == 'ch1' or aperture == 'ch2':
            aperture_lu = "sw"
        elif aperture == 'ch3' or aperture == 'ch4':
            aperture_lu = "lw"
        elif disperser is None or disperser == 'p750l':
            aperture_lu = "imager"

        det_dict = {key: get_json_parameters(self.det_pars, key, aperture=aperture_lu)
                    for key in self.det_pars.keys() if not key == 'meta'}

        return det_dict

    def get_variance_fudge(self, wave):
        """
        In addition to a scalar fudge factor, MIRI also has a chromatic variance fudge that correlates with
        the quantum yield. The information posted in https://github.com/STScI-SSB/pandeia/issues/2167 suggests that
        they want the noise scaled by an extra factor of the quantum yield so that the SNR scales inversely
        with quantum yield. The MIRI team has been asked to provide the chromatic fudge factor they want as a separate
        reference file. Until that's delivered, we'll use the quantum yield squared to achieve the desired effect.
        """
        det_pars = self.get_detector_pars()
        scalar_fudge = det_pars.get('var_fudge', 1.0)
        q_yield, fano_factor = self.get_quantum_yield(wave)
        var_fudge = scalar_fudge * q_yield**2

        return var_fudge

    def get_dichroic(self, type, level):
        """
        MIRI dichroic reference files are keyed by type (refl or trans), level, and disperser

        Parameters
        ----------
        type: str
            Either 'refl' for reflective or 'trans' for transmitting
        level: int
            Level 1, 2, or 3

        Returns
        -------
        key: str
            Key pointing to the dichroic efficiency data in self.paths
        """
        disperser = self.instrument['disperser']
        if disperser == 'short':
            mrs_setting = 'a'
        elif disperser == 'medium':
            mrs_setting = 'b'
        elif disperser == 'long':
            mrs_setting = 'c'

        key = "sd" + str(level) + mrs_setting + "_" + type
        return key

    def get_internal_eff(self, wave):
        """
        Calculate MIRI internal efficiency which is rather more complicated than the other instruments

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate efficiency onto

        Returns
        -------
        eff: numpy.ndarray or float
            Internal efficiency as a function of wave
        """
        aperture = self.instrument['aperture']
        mirror_eff = self.mirror_eff
        mirror_cont = self.mirror_cont
        n_refl = self.n_reflections[aperture]
        if self.mode == 'mrs':
            if aperture == 'ch1':
                d1r_key = self.get_dichroic('refl', 1)
                d1r_eff = self._get_throughput(wave, d1r_key)
                refl_eff = mirror_eff ** n_refl
                internal_eff = d1r_eff * refl_eff

            elif aperture == 'ch2':
                d1t_key = self.get_dichroic('trans', 1)
                d2r_key = self.get_dichroic('refl', 2)
                d1t_eff = self._get_throughput(wave, d1t_key)
                d2r_eff = self._get_throughput(wave, d2r_key)
                refl_eff = mirror_eff ** n_refl
                internal_eff = d1t_eff * d2r_eff * refl_eff

            elif aperture == 'ch3':
                d1t_key = self.get_dichroic('trans', 1)
                d2t_key = self.get_dichroic('trans', 2)
                d3r_key = self.get_dichroic('refl', 3)
                d1t_eff = self._get_throughput(wave, d1t_key)
                d2t_eff = self._get_throughput(wave, d2t_key)
                d3r_eff = self._get_throughput(wave, d3r_key)
                refl_eff = mirror_eff ** n_refl
                internal_eff = d1t_eff * d2t_eff * d3r_eff * refl_eff

            elif aperture == 'ch4':
                d1t_key = self.get_dichroic('trans', 1)
                d2t_key = self.get_dichroic('trans', 2)
                d3t_key = self.get_dichroic('trans', 3)
                d1t_eff = self._get_throughput(wave, d1t_key)
                d2t_eff = self._get_throughput(wave, d2t_key)
                d3t_eff = self._get_throughput(wave, d3t_key)
                refl_eff = mirror_eff ** n_refl
                internal_eff = d1t_eff * d2t_eff * d3t_eff * refl_eff

            else:
                raise EngineInputError(value='Invalid aperture for MIRI: %s' % aperture)

        else:
            refl_eff = mirror_eff ** n_refl
            internal_eff = wave * 0.0 + refl_eff

        # mirror contamination factor
        internal_eff = internal_eff * mirror_cont

        return internal_eff

    def get_disperser_eff(self, wave):
        """
        Overloaded here because disperser efficiency is keyed off of the aperture rather than disperser

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto

        Returns
        -------
        disperser_eff: numpy.ndarray or float
            Disperser efficiency as a function of wave
        """
        if self.instrument['disperser'] is not None:
            key = self.instrument['aperture']
            disperser_eff = self._get_throughput(wave, key)
        else:
            disperser_eff = 1.
        return disperser_eff

    def _get_dispersion_key(self):
        """
        Overload this because the key is constructed from both the aperture and disperser rather
        than the disperser alone.

        Returns
        -------
        key: str
            Key used to get dispersion file out of self.paths
        """
        disperser = self.instrument['disperser']
        aperture = self.instrument['aperture']
        key = "%s_%s_disp" % (aperture, disperser)
        return key



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, absolute_import

"""
These are the wavelength and flux units that are used internally for all calculations.
"""
pandeia_waveunits = "micron"
pandeia_fluxunits = "mjy"

"""
Absolute and relative tolerance values to use to check if values are close enough to be considered equal.
See https://docs.scipy.org/doc/numpy/reference/generated/numpy.isclose.html for details on how they're applied
in comparisons.
"""
pandeia_atol = 1.0e-8
pandeia_rtol = 1.0e-6



from distutils import log
from distutils.command.build_ext import build_ext
from distutils.errors import DistutilsError, CCompilerError, CompileError
from distutils.util import strtobool

from configparser import ConfigParser


class build_optional_ext(build_ext):
    """This is a version of the build_ext command that allows specific
    extensions to be marked as 'optional'.  If an optional extension fails to
    build, a warning message is displayed but the build does not cancel.

    It should be noted that this functionality already exists in the Python3
    versions of distutils and packaging, but we must provide backwards
    compatibility for it.
    """

    command_name = 'build_ext'

    def _find_optional_extensions(self, extensions):
        """Reads the setup.cfg to determine which extensions were set as
        'optional'.  Optionally, developers may also provide a warning message
        to display (otherwise a default message is used).  This is one way in
        which this command improves on the existing functionality on Python 3.
        """

        # TODO: However, support for the 'optional' extension attribute should
        # be included in d2to1, since it's a legitimate backwards compatibility
        # issue.  But until I do another release it will have to be supported
        # here.

        cfg = ConfigParser()
        try:
            cfg.read('setup.cfg')
        except Exception as e:
            log.warn('Failed to read setup.cfg: %s; proceeding as though '
                     'there are no optional extensions' % e)
            return

        # Map extension names to extension objects
        extensions = dict((ext.name, ext) for ext in extensions)

        for section in cfg.sections():
            if not section.startswith('extension='):
                continue

            # The extension name can be specified with the 'name' option, but
            # if that's missing the name is taken from the section header for
            # now
            if cfg.has_option(section, 'name'):
                name = cfg.get(section, 'name')
            else:
                _, name = section.split('=', 1)

            if name not in extensions:
                # Could happen, for example, if a setup_hook disabled some
                # extensions
                continue

            ext = extensions[name]

            if cfg.has_option(section, 'optional'):
                ext._optional = strtobool(cfg.get(section, 'optional'))
            else:
                ext._optional = False

            if cfg.has_option(section, 'fail_message'):
                ext._fail_message = cfg.get(section, 'fail_message')

    def check_extensions_list(self, extensions):
        build_ext.check_extensions_list(self, extensions)
        self._find_optional_extensions(extensions)

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except (CCompilerError, DistutilsError, CompileError) as e:
            if not hasattr(ext, '_optional') or not ext._optional:
                raise
            log.warn('building optional extension "%s" failed: %s' %
                     (ext.name, e))
            if hasattr(ext, '_fail_message'):
                log.warn(ext._fail_message)



"""Hooks for zest.releaser specifically for STScI software"""


import glob
import os
import shutil
import sys

from configparser import ConfigParser

from setuptools.dist import Distribution
from zest.releaser.utils import ask


DEFAULT_PACKAGE_INDEX_PATH = '/eng/ssb/web/download/packages'
PACKAGE_INDEX_URL = 'http://stsdas.stsci.edu/download/packages/index'


def is_stsci_project(workingdir):
    """
    Returns True if the product being released is from STScI and is using the
    d2to1 + stsci.distutils build/install platform.

    This is determined via some basic introspection of the project layout;
    namely that it contains a setup.cfg, and that the author-email value
    contains '@stsci.edu'.  It's ham-fisted but it should do for now.
    """

    setup_cfg = os.path.join(workingdir, 'setup.cfg')
    if not os.path.exists(setup_cfg):
        return False

    cfg = ConfigParser()
    cfg.read(setup_cfg)
    if cfg.has_option('metadata', 'author-email'):
        author_email = cfg.get('metadata', 'author-email')
    elif cfg.has_option('metadata', 'author_email'):
        author_email = cfg.get('metadata', 'author_email')
    else:
        author_email = ''

    return '@stsci.edu' in author_email


def fix_dev_version_template(data):
    """
    A postreleaser.before hook to change the dev_version_template from the
    annoying default of 'x.y.z.dev0' to just 'x.y.z.dev' without the 0.
    """

    if not is_stsci_project(data['workingdir']):
        return

    data['dev_version_template'] = '%(new_version)s.dev'


def fix_sdist_format(data):
    """
    Recent versions of zest.releaser have an annoyance that it creates .zip
    sdists instead of .tar.gz.  This is supposedly to work around a bug in
    Python 2.4 with .tar.gz sdists, but none of our software supports Python
    2.4 anyways.

    Unfortunately the only way to disable this behavior, for now, is with
    monkey-patching zest.releaser.
    """

    if not is_stsci_project(data['workingdir']):
        return

    from zest.releaser.release import Releaser

    def _my_sdist_options(self):
        return ''

    Releaser._sdist_options = _my_sdist_options


def add_to_stsci_package_index(data):
    """
    A releaser.after hook to copy the source distribution to STScI's local
    package index and update the index using basketweaver.
    """

    if not is_stsci_project(data['workingdir']):
        return

    if not data['tagdir'] or not os.path.exists(data['tagdir']):
        # Do nothing if a tag checkout was not performed
        return

    if not ask('Copy source package to STScI package index'):
        return

    package_path = DEFAULT_PACKAGE_INDEX_PATH
    if not os.path.exists(package_path):
        package_path = ''

    question = 'Path to package directory'
    if package_path:
        # A default exists; let the user know
        question += ' [%s]' % package_path
    question += ': '

    answer = ''
    while not answer:
        try:
            answer = input(question).strip()
            if not answer:
                if package_path:
                    # The user simple pressed enter, so use the supplied
                    # default
                    answer = package_path
                else:
                    continue
            if not os.path.exists(answer):
                print(('The supplied path %s does not exist.  Please enter a '
                       'different path or press Ctrl-C to cancel.' % answer))
            if not os.access(answer, os.W_OK):
                print(('The supplied path %s is not writeable.  Either change '
                       'the permissions of the directory or have someone '
                       'grant you access and try again, enter a different '
                       'directory, or press Ctrl-C to cancel.' % answer))
            package_path = answer
            break
            # The default was not supplied, so keep asking
        except KeyboardInterrupt:
            return

    # A tag checkout was made and an sdist created, this is where it would be
    # (the sdist is a .zip on Windows, a .tar.gz elsewhere--normally this
    # should be .tar.gz; don't make releases on Windows)
    sdist_file = ''
    while not sdist_file:
        try:
            sdist_file = glob.glob(os.path.join(data['tagdir'], 'dist',
                                                '*.tar.gz'))[0]
        except IndexError:
            try:
                sdist_file = glob.glob(os.path.join(data['tagdir'], 'dist',
                                                    '*.zip'))[0]
            except IndexError:
                try:
                    print((
                        "Could not find a source distribution in %s; did you "
                        "do a source checkout for upload?  If possible, try "
                        "to cd to %s and manually create a source "
                        "distribution by running `python setup.py sdist`.  "
                        "Then press enter to try again (or hit Ctrl-C to "
                        "cancel).  Go ahead, I'll wait..." %
                        (data['tagdir'], data['tagdir'])))
                    input()
                except KeyboardInterrupt:
                    return

    # Almost ready go to--now we just need to check if basketweaver is
    # available, and get it if not.
    try:
        import basketweaver.makeindex
    except ImportError:
        # Use setuptools' machinery to fetch a package and add it to the path;
        # we could do this without using setuptools directly, but it would
        # basically end up reimplementing much of the same code.
        dist = Distribution({'dependency_links': [PACKAGE_INDEX_URL]})
        try:
            dist.fetch_build_eggs(['basketweaver'])
        except:
            # There are so many things that could possibly go wrong here...
            print(('Failed to get basketweaver, which is required to rebuild '
                   'the package index.  To manually complete the release, '
                   'install basketweaver manually, then copy %s into %s, cd '
                   'to %s, and then run `makeindex *`, where makeindex is the '
                   'command installed by basketweaver.' %
                   (sdist_file, package_path, package_path)))
        import basketweaver.makeindex

    # Now we should have everything we need...
    shutil.copy(sdist_file, package_path)
    old_cwd = os.getcwd()
    os.chdir(package_path)
    try:
        basketweaver.makeindex.main(glob.glob('*'))
    finally:
        os.chdir(old_cwd)

    print('Finished adding package to %s.' % PACKAGE_INDEX_URL)



import os
import shutil
import subprocess
import sys
import tempfile

import nose

from .util import reload, rmtree


TESTPACKAGE_URL = ('https://svn.stsci.edu/svn/ssb/stsci_python/'
                   'stsci.distutils/trunk/stsci/distutils/tests/testpackage')
TESTPACKAGE_REV = '17597'  # The last known 'good' revision of this package


class StsciDistutilsTestCase(object):
    @classmethod
    def setup_class(cls):
        cls.wc_dir = tempfile.mkdtemp(prefix='stsci-distutils-test-')
        try:
            p = subprocess.Popen(['svn', '-r', TESTPACKAGE_REV,
                                  '--non-interactive', '--trust-server-cert',
                                  'checkout', TESTPACKAGE_URL, cls.wc_dir],
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE)
        except OSError, e:
            raise nose.SkipTest('svn unavailable to checkout out test '
                                'package: %s' % e)

        if p.wait() != 0:
            raise nose.SkipTest('svn failed to check out the test package: '
                                '%s; tests will not be able to run' %
                                p.stderr.read().decode('latin1'))

    @classmethod
    def teardown_class(cls):
        rmtree(cls.wc_dir)

    def setup(self):
        self.temp_dir = tempfile.mkdtemp(prefix='stsci-distutils-test-')
        self.package_dir = os.path.join(self.temp_dir, 'testpackage')
        shutil.copytree(self.wc_dir, self.package_dir)
        self.oldcwd = os.getcwd()
        os.chdir(self.package_dir)

        # We need to manually add the test package's path to the stsci
        # package's __path__ since it's already been imported.
        if 'stsci' in sys.modules:
            # Clean the existing __path__ up
            reload(sys.modules['stsci'])
            sys.modules['stsci'].__path__.insert(
                0, os.path.join(self.package_dir, 'stsci'))

    def teardown(self):
        os.chdir(self.oldcwd)
        # Remove stsci.testpackage from sys.modules so that it can be freshly
        # re-imported by the next test
        for k in list(sys.modules):
            if k == 'stsci.testpackage' or k.startswith('stsci.testpackage.'):
                del sys.modules[k]
        rmtree(self.temp_dir)

    def run_setup(self, *args):
        return self._run_cmd(sys.executable, ('setup.py',) + args)

    def run_svn(self, *args):
        return self._run_cmd('svn', args)

    def _run_cmd(self, cmd, args):
        """
        Runs a command, with the given argument list, in the root of the test
        working copy--returns the stdout and stderr streams and the exit code
        from the subprocess.
        """

        os.chdir(self.package_dir)
        p = subprocess.Popen([cmd] + list(args), stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)

        streams = tuple(s.decode('latin1').strip() for s in p.communicate())

        return (streams) + (p.returncode,)





import datetime
import glob
import os
import string
import sys

from distutils import log
import imp


try:
    from packaging.util import split_multiline
except ImportError:
    try:
        from distutils2.util import split_multiline
    except ImportError:
        from d2to1.util import split_multiline


try:
    # python 2
    reload
except NameError:
    # python 3
    from imp import reload

from .svnutils import get_svn_info, get_svn_version
from .versionutils import (package_uses_version_py, clean_version_py,
                           update_setup_datetime, VERSION_PY_TEMPLATE)

# For each version.py that we create, we will note the version of
# stsci.distutils (this package) that created it.  The problem is
# when the package being installed is stsci.distutils -- we don't
# know the version yet.  So, if we can't import the version (because
# it does not exist yet), we declare it to be None and special case
# it later.
try :
    from . import version as my_version
except ImportError :
    my_version = None


def is_display_option(ignore=None):
    """A hack to test if one of the arguments passed to setup.py is a display
    argument that should just display a value and exit.  If so, don't bother
    running this hook (this capability really ought to be included with
    distutils2).

    Optionally, ignore may contain a list of display options to ignore in this
    check.  Each option in the ignore list must contain the correct number of
    dashes.
    """

    from setuptools.dist import Distribution

    # If there were no arguments in argv (aside from the script name) then this
    # is an implied display opt
    if len(sys.argv) < 2:
        return True

    display_opts = ['--command-packages', '--help', '-h']

    for opt in Distribution.display_options:
        display_opts.append('--' + opt[0])

    for arg in sys.argv:
        if arg in display_opts and arg not in ignore:
            return True

    return False


# TODO: With luck this can go away soon--packaging now supports adding the cwd
# to sys.path for running setup_hooks.  But it also needs to support adding
# packages_root.  Also, it currently does not support adding cwd/packages_root
# to sys.path for pre/post-command hooks, so that needs to be fixed.
def use_packages_root(config):
    """
    Adds the path specified by the 'packages_root' option, or the current path
    if 'packages_root' is not specified, to sys.path.  This is particularly
    useful, for example, to run setup_hooks or add custom commands that are in
    your package's source tree.

    Use this when the root of your package source tree is not in
    the same directory with the setup.py

    Config Usage::

        [files]
        packages_root = lib
        # for installing pkgname from lib/pkgname/*.py

        [global]
        setup_hooks = stsci.distutils.hooks.use_packages_root

    """

    if 'files' in config and 'packages_root' in config['files']:
        root = config['files']['packages_root']
    else:
        root = ''

    if root not in sys.path:
        if root and sys.path[0] == '':
            sys.path.insert(1, root)
        else:
            sys.path.insert(0, root)

    # Reload the stsci namespace package in case any new paths can be added to
    # it from the new sys.path entry; depending on how the namespace packages
    # are installed this may fail--specifically if it's using the old
    # setuptools-based .pth approach there is not importable package called
    # 'stsci' that can be imported by itself.
    if 'stsci' in sys.modules:
        mod = sys.modules['stsci']
        if sys.version_info[:2] >= (3, 3) and not hasattr(mod, '__loader__'):
            # Workaround for Python bug #17099 on Python 3.3, where reload()
            # crashes if a module doesn't have an __loader__ attribute
            del sys.modules['stsci']
            try:
                import stsci
            except ImportError:
                pass
        else:
            try :
                imp.reload(sys.modules['stsci'])
            except ImportError:
                # doesn't seem to bother anything when this reload() fails
                pass


def tag_svn_revision(config):
    """
    A setup_hook to add the SVN revision of the current working copy path to
    the package version string, but only if the version ends in .dev.

    For example, ``mypackage-1.0.dev`` becomes ``mypackage-1.0.dev1234``.  This
    is in accordance with the version string format standardized by PEP 386.

    This should be used as a replacement for the ``tag_svn_revision`` option to
    the egg_info command.  This hook is more compatible with
    packaging/distutils2, which does not include any VCS support.  This hook is
    also more flexible in that it turns the revision number on/off depending on
    the presence of ``.dev`` in the version string, so that it's not
    automatically added to the version in final releases.

    This hook does require the ``svnversion`` command to be available in order
    to work.  It does not examine the working copy metadata directly.


    Config Usage::

        [global]
        setup_hooks = stsci.distutils.hooks.tag_svn_revision

    You should write exactly this in your package's ``__init__.py``::

        from .version import *

    """

    if 'metadata' in config and 'version' in config['metadata']:
        metadata = config['metadata']
        version = metadata['version']

        # Don't add an svn revision unless the version ends with .dev
        if not version.endswith('.dev'):
            return

        # First try to get the revision by checking for it in an existing
        # .version module
        package_dir = config.get('files', {}).get('packages_root', '')
        packages = config.get('files', {}).get('packages', '')
        packages = split_multiline(packages)
        rev = None
        for package in packages:
            version_py = package_uses_version_py(package_dir, package)
            if not version_py:
                continue
            try:
                mod = __import__(package + '.version',
                                 fromlist='__svn_revision__')
            except ImportError:
                mod = None
            if mod is not None and hasattr(mod, '__svn_revision__'):
                rev = mod.__svn_revision__
                break

        # Cleanup
        names = set([package, package + '.'])
        for modname in list(sys.modules):
            if modname == package or modname.startswith(package + '.'):
                del sys.modules[modname]

        if rev is None:
            # A .version module didn't exist or was incomplete; try calling
            # svnversion directly
            rev = get_svn_version()

        if not rev:
            return
        if ':' in rev:
            rev, _ = rev.split(':', 1)
        while rev and rev[-1] not in string.digits:
            rev = rev[:-1]
        try:
            rev = int(rev)
        except (ValueError, TypeError):
            return

        metadata['version'] ='%s%d' % (version, rev)


def _version_hook(function_name, package_dir, packages, name, version, vdate):
    """This command hook creates an version.py file in each package that
    requires it.  This is by determining if the package's ``__init__.py`` tries
    to import or import from the version module.

    version.py will not be created in packages that don't use it.  It should
    only be used by the top-level package of the project.

    Don't use this function directly--instead use :func:`version_setup_hook` or
    :func:`version_pre_command_hook` which know how to retrieve the required
    metadata depending on the context they are run in.

    Not called directly from the config file.  See :func:`version_setup_hook`.

    """

    # Strip any revision info from version; that will be handled separately
    if '-' in version:
        version = version.split('-', 1)[0]

    for package in packages:
        version_py = package_uses_version_py(package_dir, package)
        if not version_py:
            continue

        rev = get_svn_version()
        if ((not rev or not rev[0] in string.digits) and
                os.path.exists(version_py)):
            # If were unable to determine an SVN revision and the version.py
            # already exists, just update the __setup_datetime__ and leave the
            # rest of the file untouched
            update_setup_datetime(version_py)
            return
        elif rev is None:
            rev = 'Unable to determine SVN revision'

        svn_info = get_svn_info()

        # Wrap version, rev, and svn_info in str() to ensure that Python 2
        # unicode literals aren't written, which will break things in Python 3
        template_variables = {
                'hook_function': function_name,
                'name': name,
                'version': str(version),
                'vdate': str(vdate),
                'svn_revision': str(rev),
                'svn_full_info': str(svn_info),
                'setup_datetime': datetime.datetime.now(),
        }

        # my_version is version.py for the stsci.distutils package.
        # It can be None if we are called during the install of
        # stsci.distutils; we are creating the version.py, so it was
        # not available to import yet.  If this is what is happening,
        # we construct it specially.
        if my_version is None :
            if  package == 'stsci.distutils' :
                template_variables['stsci_distutils_version'] = version
            else:
                # It should never happen that version.py does not
                # exist when we are installing any other package.
                raise RuntimeError('Internal consistency error')
        else :
            template_variables['stsci_distutils_version'] = \
                    my_version.__version__

        with open(version_py, 'w') as f:
            f.write(VERSION_PY_TEMPLATE % template_variables)


def version_setup_hook(config):
    """Creates a Python module called version.py which contains these variables:

    * ``__version__`` (the release version)
    * ``__svn_revision__`` (the SVN revision info as returned by the ``svnversion``
      command)
    * ``__svn_full_info__`` (as returned by the ``svn info`` command)
    * ``__setup_datetime__`` (the date and time that setup.py was last run).
    * ``__vdate__`` (the release date)

    These variables can be imported in the package's ``__init__.py`` for
    degugging purposes.  The version.py module will *only* be created in a
    package that imports from the version module in its ``__init__.py``.  It
    should be noted that this is generally preferable to writing these
    variables directly into ``__init__.py``, since this provides more control
    and is less likely to unexpectedly break things in ``__init__.py``.

    Config Usage::

        [global]
        setup-hooks = stsci.distutils.hooks.version_setup_hook

    You should write exactly this in your package's ``__init__.py``::

        from .version import *

    """

    if is_display_option(ignore=['--version']):
        return

    name = config['metadata'].get('name')
    version = config['metadata'].get('version', '0.0.0')
    vdate = config['metadata'].get('vdate', 'unspecified')
    package_dir = config.get('files', {}).get('packages_root', '')
    packages = config.get('files', {}).get('packages', '')

    packages = split_multiline(packages)

    _version_hook(__name__ + '.version_setup_hook', package_dir, packages,
                 name, version, vdate)


def version_pre_command_hook(command_obj):
    """
    .. deprecated:: 0.3
        Use :func:`version_setup_hook` instead; it's generally safer to
        check/update the version.py module on every setup.py run instead of on
        specific commands.

    This command hook creates an version.py file in each package that requires
    it.  This is by determining if the package's ``__init__.py`` tries to
    import or import from the version module.

    version.py will not be created in packages that don't use it.  It should
    only be used by the top-level package of the project.
    """

    if is_display_option():
        return

    package_dir = command_obj.distribution.package_dir.get('', '.')
    packages = command_obj.distribution.packages
    name = command_obj.distribution.metadata.name
    version = command_obj.distribution.metadata.version

    _version_hook(__name__ + '.version_pre_command_hook',package_dir, packages,
                 name, version, vdate=None)


def version_post_command_hook(command_obj):
    """
    .. deprecated:: 0.3
        This hook was meant to complement :func:`version_pre_command_hook`,
        also deprecated.

    Cleans up a previously generated version.py in order to avoid
    clutter.

    Only removes the file if we're in an SVN working copy and the file is not
    already under version control.
    """

    package_dir = command_obj.distribution.package_dir.get('', '.')
    packages = command_obj.distribution.packages

    for package in packages:
        clean_version_py(package_dir, package)


def numpy_extension_hook(command_obj):
    """A distutils2 pre-command hook for the build_ext command needed for
    building extension modules that use NumPy.

    To use this hook, add 'numpy' to the list of include_dirs in setup.cfg
    section for an extension module.  This hook will replace 'numpy' with the
    necessary numpy header paths in the include_dirs option for that extension.

    Note: Although this function uses numpy, stsci.distutils does not depend on
    numpy.  It is up to the distribution that uses this hook to require numpy
    as a dependency.

    Config Usage::

        [extension=mypackage.extmod]
        sources =
            foo.c
            bar.c
        include_dirs = numpy

        [build_ext]
        pre-hook.numpy-extension = stsci.distutils.hooks.numpy_extension_hook
    """

    command_name = command_obj.get_command_name()
    if command_name != 'build_ext':
        log.warn('%s is meant to be used with the build_ext command only; '
                 'it is not for use with the %s command.' %
                 (__name__, command_name))
    try:
        import numpy
    except ImportError:
        # It's virtually impossible to automatically install numpy through
        # setuptools; I've tried.  It's not pretty.
        # Besides, we don't want users complaining that our software doesn't
        # work just because numpy didn't build on their system.
        sys.stderr.write('\n\nNumpy is required to build this package.\n'
                         'Please install Numpy on your system first.\n\n')
        sys.exit(1)

    includes = [numpy.get_include()]
    #includes = [numpy.get_numarray_include(), numpy.get_include()]
    for extension in command_obj.extensions:
        if 'numpy' not in extension.include_dirs:
            continue
        idx = extension.include_dirs.index('numpy')
        for inc in includes:
            extension.include_dirs.insert(idx, inc)
        extension.include_dirs.remove('numpy')


def glob_data_files(command_obj):
    """A pre-command hook for the install_data command allowing wildcard
    patterns to be used in the data_files option.

    Also ensures that data files with relative paths as their targets are
    installed relative install_lib.

    Config Usage::

        [files]
        data_files =
            target_directory = source_directory/*.foo
            other_target_directory = other_source_directory/*

        [install_data]
        pre-hook.glob-data-files = stsci.distutils.hooks.glob_data_files


    """

    command_name = command_obj.get_command_name()
    if command_name != 'install_data':
        log.warn('%s is meant to be used with the install_data command only; '
                 'it is not for use with the %s command.' %
                 (__name__, command_name))

    data_files = command_obj.data_files

    for idx, val in enumerate(data_files[:]):
        if isinstance(val, str):
            # Support the rare, deprecated case where just a filename is given
            filenames = glob.glob(val)
            del data_files[idx]
            data_files.extend(filenames)
            continue

        dest, filenames = val
        filenames = sum((glob.glob(item) for item in filenames), [])
        data_files[idx] = (dest, filenames)

    # Ensure the correct install dir; this is the default behavior for
    # installing with distribute, but when using
    # --single-version-externally-managed we need to to tweak this
    install_cmd = command_obj.get_finalized_command('install')
    if command_obj.install_dir == install_cmd.install_data:
        install_lib_cmd = command_obj.get_finalized_command('install_lib')
        command_obj.install_dir = install_lib_cmd.install_dir



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
A set of exceptions for pyetc to raise

The web interface will recognize any of these exceptions and produce a
pretty and user-friendly (to the extent that you gave good text) error
message.

"""
from __future__ import division, absolute_import


class PyetcException(Exception):

    """superclass of all custom pyetc exceptions

    This exception contains these fields of interest:

        message - the message provided by the code that raised the exception

        results - the results dict that the engine was working on at the
            time of the exception (this may be an empty dictionary if the
            engine code cannot provide it)

    """

    def __init__(self, value='Unspecified Error', results=None):
        Exception.__init__(self, value)
        self.results = results
        self.name = "Unspecified pyetc exception"

    def addinfo(self, more_info):
        """ A convenience method so a calling function can add additional
        information to a raised exception before reraising.
        """
        args = list(self.args)
        args[0] = str(args[0]) + " " + str(more_info)
        self.args = tuple(args)

        # self.message is deprecated as of python 2.6, but we must set it
        # now in order to get the desired behavior.
        self.message = self.args[0]


class EngineInputError(PyetcException):

    """
    an engine error that is probably related to user input

    raise engine.exceptions.EngineInputError( 'descriptive text', result_dict )

    The result dict is the partially completed result dictionary that the
    engine would have returned if it could finish the calculation.

    This error should include the etcid that issued it

    """

    def __init__(self, value="Input Error", results=None):
        PyetcException.__init__(self, value)
        self.name = "Input Error"


class DataConfigurationError(PyetcException):

    """
    an engine error that is probably related to configuration data

    raise engine.exceptions.DataConfigurationError( 'descriptive text', result_dict )

    The result dict is the partially completed result dictionary that the
    engine would have returned if it could finish the calculation.

    This error should include the etcid that issued it

    """

    def __init__(self, value="Data Configuration Error", results=None):
        PyetcException.__init__(self, value)
        self.name = "Data Configuration Error"


class EngineOutputError(PyetcException):

    """
    an engine error that is probably related to engine output

    raise engine.exceptions.OutputError( 'descriptive text', result_dict )

    The result dict is the partially completed result dictionary that the
    engine would have returned if it could finish the calculation.

    This error should include the etcid that issued it

    """

    def __init__(self, value="Output Error", results=None):
        PyetcException.__init__(self, value)
        self.name = "Output Error"


class InternalError(PyetcException):

    """'
    an error that should not have happened and is probably a bug'
    We should include the etcid in the error information?
    """

    def __init__(self, value="Input Error", results=None):
        PyetcException.__init__(self, value)
        self.name = "Internal Error"


class DataError(PyetcException):

    """'
    Some data entity which should exist cannot be found
    """

    def __init__(self, value="Data Error", results=None):
        PyetcException.__init__(self, value)
        self.name = "Data Error"


class BadEnumeration(PyetcException):

    """
    use this when an input field is expected to be one of a list of
    known values, but it is not.
    """

    def __init__(self, value=None, results=None, field=None):
        PyetcException.__init__(self, value, results)
        self.name = "Unrecognized Option"


# Make new exceptions here, if necessary.  Subclasses of InputError
# and InternalError require nothing special.  New subclasses of
# PyetcException require new code in etc_web/etc/views.py

class CalculationError(PyetcException):

    """ A catch-all exception for otherwise nonspecific exceptions
    that are raised by the engine.
    """

    def __init__(self, value=None, results=None):
        PyetcException.__init__(self, value, results)
        self.name = "Calculation Error"


class PysynphotError(PyetcException):

    """ Any problem that arises from pysynphot should raise
    this exception
    """

    def __init__(self, value, results=None):
        PyetcException.__init__(self, value, results)
        self.name = "Pysynphot Error"


class RangeError(PyetcException):

    """ Any problem caused by some value (usually a wavelength) falling
    outside a valid range should raise this exception.
    """

    def __init__(self, value, results=None):
        PyetcException.__init__(self, value, results)
        self.name = "Value Out of Range"


class UnsupportedError(PyetcException):

    """ Certain cases are 'available' but 'unsupported'.
    Raise this exception for those cases.
    """

    def __init__(self, value, results=None):
        PyetcException.__init__(self, value, results)
        self.name = "Unsupported Option Selected"


class EEDataNotAvailable(PyetcException):

    """ EEData refers to data from the Enclosed Energy table provided
    for a given instrument, used for point sources.
    If no data is available to support a requested operation, this
    exception is raised. The typical case is one in which the Enclosed
    Energy table does not extend far enough in wavelength.
    """

    def __init__(self, value, results=None):
        PyetcException.__init__(self, value, results)
        self.name = "Enclosed Energy data not available"


# Consistency checking exceptions used by Collections
class InstrumentMismatch(PyetcException):

    """ Raised when attempting to combine products of different
    instruments or differently configured instruments.
    """

    def __init__(self, value, results=None):
        PyetcException.__init__(self, value, results)
        self.name = "Inconsistent instruments"


class WavesetMismatch(PyetcException):

    """ Raised when attempting to combine spectra of different
    wavelength sets
    """

    def __init__(self, value, results=None):
        PyetcException.__init__(self, value, results)
        self.name = "Inconsistent wavelength arrays"


class ShapeMismatch(PyetcException):

    """
    Raised when Shapes that should match don't.

    """

    def __init__(self, value, results=None):
        PyetcException.__init__(self, value, results)
        self.name = "Shapes are not the same"


class ExtractionMismatch(PyetcException):

    """
    Raised when Extracted instances that should match don't.

    """

    def __init__(self, value, results=None):
        PyetcException.__init__(self, value, results)
        self.name = "Extracted instances are not the same"


# ### BACKGROUND ERRORS ### #

class BackgroundError(PyetcException):

    """
    Raised when the Background Butler runs into errors.

    These errors could include:
        - invalid (position, date) combination
        - no stray light data available

    Usage:
        raise engine.custom_exceptions.BackgroundError('descriptive text')
    """

    def __init__(self, value='Background Error', debug=None):
        PyetcException.__init__(self, value)
        self.name = value
        if debug:
            print('Background Error:')
            print(debug)


class StraylightPositionError(BackgroundError):

    """
    Raised when the specified position (ra, dec) is not observable on the
    specified date.

    Usage:
        raise engine.custom_exceptions.StraylightPositionError('descriptive text')
    """

    def __init__(self, value='Background Error', debug=None):
        BackgroundError.__init__(self, value, debug)
        self.name = value


class StraylightDataError(BackgroundError):

    """
    Raised when needed stray light data is missing.

    Usage:
        raise engine.custom_exceptions.StraylightDataError('descriptive text')
    """

    def __init__(self, value='Background Error', debug=None):
        BackgroundError.__init__(self, value, debug)
        self.name = value


class DatelessBGError(BackgroundError):

    def __init__(self, value='Dateless Background Error', debug=None):
        BackgroundError.__init__(self, value, debug)
        self.name = value


class BMGError(BackgroundError):

    """
    Raised when the BMG returns an unhappy status code.

    Usage:
        raise engine.custom_exceptions.BMGError('status code')
    """

    def __init__(self, status, ra_dec_str, date_str):

        # these are the codes we can share with the user
        public_codes = {
            '500': 'BMG server error.',
            '501': 'Date (%s) out of range.' % date_str,
            '502': 'Ra/Dec (%s) out of range.' % ra_dec_str,
            '503': 'Wavelength out of range.'
        }

        # these are codes that we will log but won't share with the user
        private_codes = {
            '400': 'Empty bmg block (e.g. nothing after bmg keyword)',
            '401': 'No bmg block (e.g. no bmg keyword)',
            '402': 'Bad query block (e.g. badly formatted block)',
            '403': 'Bad query_type value (e.g. miss-spelled or un-supported)',
            '404': 'Empty query_type value (e.g. missing query_type value)',
            '405': 'No query_type keyword in query block (e.g. miss-spelled)',
            '504': 'Kernel list is inaccessible',
            '505': 'File wainscoat.txt is inaccessible',
            '506': 'Schlegel interstellar dust map fits file not found or inaccessible',
            '507': 'Schlegel interstellar dust map file name too long',
            '508': 'Error reading the std_spectrum_wavelengths.txt file',
            '509': 'Wavelength outside (0.1, 1200) micros for CIB',
            '510': 'Cib_spec_mode returns -1, wl outside (0.1, 1200)',
            '511': 'Error in opening file starcount.txt (when environment variable OUTPUT_STARCOUNT is set)'
        }

        if status in public_codes:
            # pass the error message to the user
            BackgroundError.__init__(self, public_codes[status])
        elif status in private_codes:
            # log the actual error, but just show the user something generic
            msg = 'The background model server encountered an unexpected error. TODO: insert generic error message text here.'
            BackgroundError.__init__(self, msg, private_codes[status])
        else:
            # Unexpected/unhandled error, but pass it on
            BackgroundError.__init__(self, "BG error with unknown status of: "+str(status))



"""AST Visitors

Currently only uses one for collecting a list of import statements.
Unfortunately two versions of this have to be implemented: One for 2.6 and up
and a different version for 2.5.
"""


import os
import sys

from distutils import log

try:
    import ast # Python >= 2.6

    def walk(filename, visitor):
        """Generate an AST for the given filename and walk over it using
        the given visitor instance.
        """

        filename = os.path.abspath(filename)

        try:
            tree = ast.parse(open(filename, 'r').read())
        except SyntaxError:
            if sys.version_info[0] < 3:
                e = sys.exc_info()[1]
                log.warn('SyntaxError while parsing file %s: %s' %
                         (filename, str(e)))
                return
            # We're probably in Python 3 and looking at a file intended for
            # Python 2.  Otherwise there's an unintended SyntaxError in the
            # file, so there are bigger problems anyways
            try:
                import lib2to3.refactor

                tool = StringRefactoringTool(
                    lib2to3.refactor.get_fixers_from_package('lib2to3.fixes'))
                tool.refactor_file(filename, write=True)
                tree = ast.parse(tool.refactored[filename])
            except ImportError:
                # Without 2to3 we can't do much more.
                # TODO: Issue a warning?
                return

        visitor.visit(tree)


    class ImportVisitor(ast.NodeVisitor):
        def __init__(self):
            self.imports = set()
            self.importfroms = set()

        def visit_Import(self, node):
            for name in node.names:
                self.imports.add((name.name, name.asname))

        def visit_ImportFrom(self, node):
            for name in node.names:
                self.importfroms.add((node.module, name.name, name.asname))

except ImportError:
    import compiler

    def walk(filename, visitor):
        tree = compiler.parseFile(filename)
        compiler.walk(tree, visitor)

    class ImportVisitor(compiler.visitor.ASTVisitor):
        def __init__(self):
            self.imports = set()
            self.importfroms = set()

        def visitImport(self, node):
            for name in node.names:
                self.imports.add(name)

        def visitFrom(self, node):
            for name in node.names:
                self.importfroms.add((node.modname, name[0], name[1]))


if sys.version_info[0] >= 3:
    try:
        import lib2to3.refactor

        class StringRefactoringTool(lib2to3.refactor.RefactoringTool):
            """A RefactoringTool that saves refactored files as strings in the
            self.refactored dict rather than outputting to actual files.

            This is used in case we're running in Python 3 and need to refactor
            a file before parsing its syntax tree.
            """

            def __init__(self, fixer_names, options=None, explicit=None):
                super(StringRefactoringTool, self).__init__(fixer_names,
                                                            options,
                                                            explicit)
                self.refactored = {}

            def write_file(self, new_text, filename, old_text, encoding=None):
                self.refactored[filename] = new_text

    except ImportError:
        pass



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, absolute_import

from .telescope import Telescope
from .instrument import Instrument


class HST(Telescope):

    """
    Currently a dummy class for directory/file discovery, but could eventually contain HST-specific methods
    """
    pass


class HSTInstrument(Instrument):

    """
    Generic HST Instrument class
    """
    def __init__(self, mode=None, config={}, **kwargs):
        telescope = HST()
        Instrument.__init__(self, telescope=telescope, mode=mode, config=config, **kwargs)


class WFC3(HSTInstrument):

    """
    Currently HST WFC3 requires no extra methods beyond those provided by the generic Instrument class
    """
    pass



from __future__ import with_statement

import os
import sys

try:
    import numpy
except ImportError:
    numpy = None

from nose import SkipTest

from . import StsciDistutilsTestCase
from .util import get_compiler_command, open_config


class TestCommands(StsciDistutilsTestCase):
    def test_build_optional_ext(self):
        if numpy is None:
            raise SkipTest("numpy is required to run this test")
        # The test extension in the test package is already configured to be
        # "optional" by default--we'll do one test build to make sure that goes
        # smoothly
        compiler_cmd = get_compiler_command()

        _, _, exit_code = self.run_setup('build')

        # Make sure the build went successfully; a zero exit code should be
        # good enough for our purposes
        assert exit_code == 0

        # Now let's try breaking the build
        with open(os.path.join('src', 'testext.c'), 'a') as f:
            f.write('1/0')

        # We leave off the exit status from the compiler--in most cases it will
        # say "exit status 1" but that can't be guaranteed for all compilers
        msg = ('building optional extension "stsci.testpackage.testext" '
               'failed: command \'%s\' failed with exit status' % compiler_cmd)
        # Prior to Python 2.7, distutils.log output everything to stdout; now
        # warnings and errors are output to stderr
        if sys.version_info[:2] < (2, 7):
            stderr, _, exit_code = self.run_setup('build', '--force')
        else:
            _, stderr, exit_code = self.run_setup('build', '--force')
        assert exit_code == 0
        assert stderr.splitlines()[-1].startswith(msg)

        # Test a custom fail message
        with open_config('setup.cfg') as cfg:
            cfg.set('extension=stsci.testpackage.testext', 'fail_message',
                    'Custom fail message.')

        if sys.version_info[:2] < (2, 7):
            stderr, _, exit_code = self.run_setup('build', '--force')
        else:
            _, stderr, exit_code = self.run_setup('build', '--force')
        assert exit_code == 0
        assert stderr.splitlines()[-1] == 'Custom fail message.'

        # Finally, make sure the extension is *not* treated as optional if not
        # marked as such in the config
        with open_config('setup.cfg') as cfg:
            cfg.remove_option('extension=stsci.testpackage.testext',
                              'optional')

        # This error message comes out on stderr for all Python versions AFAICT
        msg = "error: command '%s' failed with exit status" % compiler_cmd
        _, stderr, exit_code = self.run_setup('build', '--force')
        assert exit_code != 0
        assert stderr.splitlines()[-1].startswith(msg)



#!/usr/bin/env python

# This sample setup.py can be used as a template for any project using d2to1.
# Simply copy this file and, if desired, delete all the comments.  Also remove
# the 'namespace_packages' and 'packages' arguments to setup.py if this project
# does not contain any packages beloning to a namespace package.

# This import statement attempts to import the setup() function from setuptools
# (this replaces the setup() one uses from distutils.core when using plain
# distutils).
#
# If the import fails (the user doesn't have the setuptools package) it then
# uses the ez_setup bootstrap script to install setuptools, then retries the
# import.  This is common practice for packages using setuptools.
try:
    from setuptools import setup
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup

# The standard setup() call.  Notice, however, that most of the arguments
# normally passed to setup() are absent.  They will instead be read from the
# setup.cfg file using d2to1.
#
# In order for this to work it is necessary to specify setup_requires=['d2to1']
# If the user does not have d2to1, this will boostrap it.  Also require
# stsci.distutils to use any of the common setup_hooks included in
# stsci.distutils (see the setup.cfg for more details).
#
# The next line, which defines namespace_packages and packages is only
# necessary if this projet contains a package belonging to the stsci namespace
# package, such as stsci.distutils or stsci.tools.  This is not necessary for
# projects with their own namespace, such as acstools or pyraf.
#
# d2to1=True is required to enable d2to1 and read the remaning project metadata
# from the setup.cfg file.
#
# use_2to3 and zip_safe are common options support by setuptools; these can
# also be placed in the setup.cfg, as will be demonstrated in a future update
# to this sample package.
setup(
    setup_requires=['d2to1>=0.2.11', 'stsci.distutils>=0.3.7'],
    namespace_packages=['pandeia'], packages=['pandeia'],
    d2to1=True
)




# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, absolute_import

import numpy as np
import numpy.ma as ma
import copy
import astropy.io.fits as fits

from .coords import Grid, IrregularGrid
from .custom_exceptions import EngineOutputError
from six.moves import range
from six.moves import zip


class Report(object):

    """
    This class contains functionality for generating reports and data products from the results
    of an ETC calculation.

    Parameters
    ----------
    calc_input: dict
        The engine API format dict used to configure the calculation
    signal_list : list of etc3D.DetectorSignal instances
        The calculated signal on the detector
    noise_list : list of etc3D.DetectorNoise instances
        The calculated noise on the detector
    extracted : dict
        The extracted data product generated by the Strategy
    warnings: dict
        Dict of warnings produced and collected through the course of the calculation
    """

    def __init__(self, calc_input, signal_list, noise_list, saturation_list, extracted, warnings):
        # For single pointing calculations, the signal and noise lists have one element.
        # However, handling calculations that involve multiple pointings is more tricky.
        # We want to grab one of the on_target calculations to get the relevant target
        # information for the report.
        if not isinstance(signal_list, list):
            raise EngineOutputError(value="Must provide Report signals and noises as lists.")

        self.extracted = extracted

        self.sub_reports = []
        if len(signal_list) == 1:
            self.signal = signal_list[0]
            self.noise = noise_list[0]
            self.saturation = saturation_list[0]
        elif len(signal_list) > 1:
            # Fish out the list of which ones are targets.
            """
            on_target = calc_input['strategy']['on_target']  # True if target, False otherwise
            target = [i for i, v in enumerate(on_target) if v][0]  # get first target
            """
            # The above method does not work in general. Assume the first plane is a target and
            # refactor when implementing dithers in general.
            target = 0

            self.signal = signal_list[target]
            self.noise = noise_list[target]
            self.saturation = saturation_list[target]
            for signal, noise, saturation in zip(signal_list, noise_list, saturation_list):
                r = Report(calc_input, [signal], [noise], [saturation], extracted, warnings)
                self.sub_reports.append(r)

        # We use the calculation inputs for a few things.
        self.input = calc_input

        # Get the full Signal and Noise objects for inspection.
        self.signals = signal_list
        self.noises = noise_list
        self.saturations = saturation_list

        # get the ExposureSpecification.
        self.exposure_specification = self.signal.current_instrument.get_exposure_pars()

        # get the type of observation. basically, this is the format of the detector plane data.
        # it will be either 'image' (X spatial, Y spatial) or 'spec' (X wavelength, Y spatial)
        self.projection_type = self.signal.projection_type

        # Target spectrum
        target_curve = [self.signal.wave, self.signal.total_flux]

        # Transmission/focal plane rate
        trans_curve = [self.signal.wave, self.signal.fp_rate.sum(axis=(0, 1))]

        # input background
        bg_curve = [self.signal.wave, self.signal.background.MJy_sr]

        # Background rate
        bg_rate_curve = [self.signal.wave, self.signal.bg_fp_rate]

        # this is the wavelength sampling of the calculation
        self.wave = self.signal.wave

        # this is the data cube of the input signal
        self.flux = self.signal.flux_cube_list

        # this is the data cube of the input signal plus background
        self.flux_plus_bg = self.signal.flux_plus_bg_list

        # This is the background rate in each pixel without sources
        self.bg_pix = self.signal.rate_plus_bg - self.signal.rate

        # Signal and noise in 2D. Get from extracted products
        s = extracted['detector_signal']
        n = extracted['detector_noise']

        # get areas in pixels of source and background regions
        self.extraction_area = extracted['extraction_area']
        if extracted['background_area'] is not None:
            self.background_area = extracted['background_area']
        else:
            self.background_area = None

        # Check for full saturation and set the noise to infinity accordingly. This is necessary since
        # the engine currently returns noise=0 if the pixel has full saturation, which is confusing since the
        # noise is not 0, but rather undertermined or infinite. Setting the noise to NaN ensures that the S/N
        # of saturated pixels are NaNs.
        n[self.saturation == 2] = np.nan
        self.detector_sn = (s - self.bg_pix) / n
        self.detector_signal = s + n * np.random.randn(n.shape[0], n.shape[1])

        # this is the spatial grid for the calculation
        self.grid = self.signal.grid

        # this is the detector plane pixel grid. for most modes we just grab and use it directly.
        # however, slitless we want to redefine it to be spatial on both axes.
        if self.signal.projection_type in ('slitless'):
            orig_grid = self.signal.pixgrid_list[0]
            # detector plane gets rotated depending on dispersion_axis so adjust accordingly
            if self.signal.dispersion_axis == 'x':
                self.pix_grid = Grid(self.grid.xsamp, orig_grid.ysamp, orig_grid.nx, orig_grid.ny)
            else:
                self.pix_grid = Grid(self.grid.ysamp, orig_grid.xsamp, orig_grid.ny, orig_grid.nx)
        elif self.input['strategy']['method'] in ('ifuapphot', 'ifunodinscene', 'ifunodoffscene'):
            # for IFUs we want the reconstructed image plane that's build by the strategy
            self.pix_grid = extracted['plane_grid']
        elif self.signal.projection_type in ('multiorder'):
            orig_grid = self.signal.pixgrid_list[0]
            self.pix_grid = IrregularGrid(np.arange(orig_grid.nx), np.arange(orig_grid.ny))
        else:
            self.pix_grid = self.signal.pixgrid_list[0]

        # Values calculated by the strategy
        sn = extracted['extracted_flux'] / extracted['extracted_noise']
        if self.signal.projection_type == 'image':
            # this is the wavelength sampling on the detector. in imaging mode this is
            # the effective wavelength of the filter + detector + optics
            self.wave_pix = self.signal.wave_pix  # convert to real np.array of len = 1
        elif self.signal.projection_type == 'spec':
            # this is the wavelength sampling on the detector.
            self.wave_pix = extracted['wavelength']  # this is already a 1D np.array
            self.cube_signal, self.cube_noise, self.cube_saturation, self.cube_plane_grid = extracted['reconstructed']
            self.cube_sim = self.cube_signal + self.cube_noise * np.random.randn(*self.cube_noise.shape)
        elif self.signal.projection_type == 'slitless':
            self.wave_pix = extracted['wavelength']
            self.detector_sn_unrot = self.detector_sn
            self.detector_signal_unrot = self.detector_signal
            self.saturation_unrot = self.saturation
            if self.signal.dispersion_axis == 'y':
                # need to rotate these 90 deg clockwise to match our normal axis orientation. np.rot90 only works CCW
                # so need to flip, rotate, and then flip back. note that currently this case implies that dispersion
                # axis is 90 degrees.
                self.wave_pix = self.wave_pix[::-1]
                self.detector_sn = np.flipud(np.rot90(np.flipud(self.detector_sn)))
                self.detector_signal = np.flipud(np.rot90(np.flipud(self.detector_signal)))
                self.saturation = np.flipud(np.rot90(np.flipud(self.saturation)))
        elif self.signal.projection_type == 'multiorder':
            self.wave_pix = extracted['wavelength']
            self.detector_sn_unrot = self.detector_sn
            self.detector_signal_unrot = self.detector_signal
            self.saturation_unrot = self.saturation
            self.detector_sn = np.rot90(self.detector_sn)
            self.detector_signal = np.rot90(self.detector_signal)
            self.saturation = np.rot90(self.saturation)
        else:
            raise EngineOutputError(value="Unsupported projection_type: %s" % self.signal.projection_type)

        sn_curve = [self.wave_pix, sn]
        extracted_noise = [self.wave_pix, extracted['extracted_noise']]
        extracted_flux = [self.wave_pix, extracted['extracted_flux']]
        extracted_flux_plus_bg = [self.wave_pix, extracted['extracted_flux_plus_bg']]
        
        total_flux = [self.wave_pix, extracted['source_flux_in_fov']]
        self.curves = {
            'target': target_curve,
            'fp': trans_curve,
            'bg': bg_curve,
            'bg_rate': bg_rate_curve,
            'sn': sn_curve,
            'extracted_noise': extracted_noise,
            'extracted_flux': extracted_flux,
            'extracted_flux_plus_bg': extracted_flux_plus_bg,
            'total_flux': total_flux,
            'extracted_bg_total': [self.wave_pix, extracted['extracted_bg_total']],
            'extracted_bg_only': [self.wave_pix, extracted['extracted_bg_only']],
            'extracted_contamination': [self.wave_pix, extracted['extracted_bg_total'] - extracted['extracted_bg_only']],
            'n_partial_saturated': [self.wave_pix,extracted['saturation_products']['partial']],
            'n_full_saturated': [self.wave_pix,extracted['saturation_products']['full']]
        }
        # If there is a contrast product, add it
        if 'contrast_curve' in extracted:
            self.curves['contrast'] = extracted['contrast_curve']

        self.warnings = warnings

    def as_dict(self):
        """
        Produce report in dictionary format conformant with the engine API.

        Most strategies operate on and extract flux, but some (eg TACentroid) use the same machinery, which is
        in principle completely general, to extract a properly weighted sum of something else (eg, centroid
        for TACentroid).  These strategies, the "extracted_flux" is the extracted product of interest, and
        the "extracted_noise" is the standard deviation of the extracted product.

        Returns
        -------
        r: dict
            Dictionary containing the results of the calculation, associated information,
            and any applicable warnings.
        """
        # fill out report with the calculation outputs...
        r = {}

        # look for any sub-reports and convert them
        r['sub_reports'] = []
        for report in self.sub_reports:
            r['sub_reports'].append(report.as_dict())

        # set up the coordinate transform information that describes the image axes
        # for the 2D and 3D data.  first, get the spatial information from self.Grid
        r['transform'] = self.pix_grid.as_dict()

        # the wavelength sampling of the input spectrum is different
        # then the sampling by the detector. the model cubes that are
        # used to make the calculation use the input sampling. the detector
        # plane outputs for spectroscopy and IFU use the detector sampling

        # ## input model wavelength sampling ## #
        r['transform'].update(self.signal.spectral_model_transform())

        # ## detector plane wavelength sampling ## #
        r['transform'].update(self.signal.spectral_detector_transform())

        # the spectral mapping can be shifted by strategies (e.g. slitless). this is
        # reflected in wave_pix so update from there.
        r['transform']['wave_det_min'] = self.wave_pix.min()
        r['transform']['wave_det_max'] = self.wave_pix.max()
        r['transform']['wave_det_refval'] = self.wave_pix[0]

        # scalar data products
        r['scalar'] = {}
        # if there's a detector gap, the noise is 0 and sn goes to np.inf. fix that here
        # so that we can get a proper maximum S/N and report scalar values there.
        # NOTE: this value is only used in the scalar values, the 1d sn
        # may still contain a NaN.
        sn = ma.fix_invalid(self.curves['sn'][1], fill_value=0.0).data

        if self.input['strategy']['method'] in [
            'soss',
            'specapphot',
            'msafullapphot',
            'ifuapphot',
            'ifunodinscene',
            'ifunodoffscene'
        ]:
            wave_index = int(len(self.wave_pix) / 2.0)
            if 'reference_wavelength' in self.input['strategy']:
                wref = self.input['strategy']['reference_wavelength']
                if wref is not None:
                    if wref >= self.wave_pix.min() and wref <= self.wave_pix.max():
                        wave_index = (np.abs(self.wave_pix - wref)).argmin()
                    else:
                        self.warnings['bad_waveref'] = "Specified wavelength, %f, out of range [%f, %f]. " % (
                            wref,
                            self.wave_pix.min(),
                            self.wave_pix.max()
                        )
                        self.warnings['bad_waveref'] += "Using %f to select diagnostic planes instead." % (
                            float(self.wave_pix[wave_index])
                        )

        if self.input['strategy']['method'] in ['imagingapphot', 'coronagraphy', 'tacentroid', 'taphot']:
            wave_index = 0

        # the 3D data products
        r['3d'] = {}
        r['3d']['flux'] = self.flux  # model flux cube
        r['3d']['flux_plus_background'] = self.flux_plus_bg  # model flux cube plus background
        if self.input['strategy']['method'] in ['ifuapphot', 'ifunodinscene', 'ifunodoffscene']:  # IFU mode generates 3D cubes.
            r['3d']['reconstructed'] = self.cube_sim
            r['3d']['reconstructed_signal'] = self.cube_signal
            r['3d']['reconstructed_noise'] = self.cube_noise
            r['3d']['reconstructed_saturation'] = self.cube_saturation
            r['3d']['reconstructed_snr'] = self.cube_signal / self.cube_noise

        # the 2D data products
        r['2d'] = {}
        if self.input['strategy']['method'] in ['ifuapphot', 'ifunodinscene', 'ifunodoffscene']:
            # use cube planes for IFUs, though this may be temporary. the actual detector image for an IFU
            # observation is a set of spectra, one for each IFU slice. populate the detector and SNR images from the
            # planes of the reconstructed cubes, but collapse the saturation cube to show the most severe saturation
            # for each spatial pixel.
            r['2d']['detector'] = self.cube_sim[wave_index, :, :]
            r['2d']['snr'] = self.cube_signal[wave_index, :, :] / self.cube_noise[wave_index, :, :]
            r['2d']['saturation'] = np.amax(self.cube_saturation, axis=0)
        else:
            r['2d']['detector'] = self.detector_signal
            r['2d']['snr'] = self.detector_sn
            r['2d']['saturation'] = self.saturation

        # make original, unrotated versions of engine 2D outputs available for slitless mode
        if self.signal.projection_type in ('slitless', 'multiorder'):
            r['2d']['detector_unrotated'] = self.detector_signal_unrot
            r['2d']['snr_unrotated'] = self.detector_sn_unrot
            r['2d']['saturation_unrotated'] = self.saturation_unrot

        # the 1D data products
        r['1d'] = {}

        # these are implicitly included with each entry in self.curves. break them out
        # explicitly as well.
        # NOTE: We do not do any NaN checking here.  This may cause certain values like SNR
        # to be reported as NaN in 1d, but as 0 in scalars.
        r['1d']['wave_pix'] = self.wave_pix
        r['1d']['wave_calc'] = self.wave
        for k in list(self.curves.keys()):
            r['1d'][k] = self.curves[k]

        rw = float(self.wave_pix[wave_index])
        r['scalar']['sn'] = float(sn[wave_index])

        # Most strategies operate on and extract flux, but some (eg TACentroid) use the same machinery, which is
        # in principle completely general, to extract a properly weighted sum of something else (eg, centroid
        # for TACentroid).  These strategies, the "extracted_flux" is the extracted product of interest, and
        # the "extracted_noise" is the standard deviation of the extracted product.

        # Some strategies may not return a flux. TargetAcqCentroid is one example for now. 
        # They are still called "extracted_flux" in the curves, but we rename them here to avoid confusion.
        # The "extracted_noise" is still the standard deviation of the extracted product.
        if self.input['strategy']['method'] in ['tacentroid']:
            r['scalar']['extracted_centroid'] = float(self.curves['extracted_flux'][1][wave_index])
            r['scalar']['extracted_flux'] = np.nan
        else:
            r['scalar']['extracted_flux'] = float(self.curves['extracted_flux'][1][wave_index])
        
        # if the SN is 0.0 (e.g. due to heavy saturation), then we must not know what the variance is so set it to np.nan
        if np.abs(r['scalar']['sn']) > 0.0:
            r['scalar']['extracted_noise'] = float(self.curves['extracted_noise'][1][wave_index])
        else:
            r['scalar']['extracted_noise'] = np.nan
        r['scalar']['background_total'] = float(self.curves['extracted_bg_total'][1][wave_index])
        r['scalar']['background_sky'] = float(self.curves['extracted_bg_only'][1][wave_index])

        # only check for contamination if background level is above some epsilon value. otherwise round-off
        # errors can cause unstable test results.
        if np.abs(r['scalar']['background_total']) > 1.0e-9:
            r['scalar']['contamination'] = (r['scalar']['background_total'] - r['scalar']['background_sky']) / \
                                            r['scalar']['background_total']
        else:
            r['scalar']['contamination'] = 0.0

        r['scalar']['reference_wavelength'] = rw
        r['scalar']['background'] = self.signal.background.bg_spec.sample(rw)
        r['scalar']['background_area'] = self.background_area
        r['scalar']['extraction_area'] = self.extraction_area

        if 'contrast_separation' in self.input['strategy']:
            contrast_separation = self.input['strategy']['contrast_separation']
            r['scalar']['contrast_separation'] = contrast_separation
            r['scalar']['contrast_azimuth'] = self.input['strategy']['contrast_azimuth']
            r['scalar']['contrast'] = np.interp(contrast_separation, self.curves['contrast'][0], self.curves['contrast'][1])

        # return some information that is not a product of a calculation
        r['information'] = {}

        # this is used by the UI to decide how to make the plots. even though IFUs are spectroscopic,
        # they're plotted as 2D spatial images. they could be shown as either single planes from the
        # reconstructed cubes or some combination of multiple planes.
        if self.input['strategy']['method'] in ['ifuapphot', 'ifunodinscene', 'ifunodoffscene']:
            r['information']['calc_type'] = "image"
        else:
            r['information']['calc_type'] = self.projection_type

        r['information']['exposure_specification'] = self.exposure_specification.__dict__

        # Some strategies like IFU and coronagraphy adds dithers. Some are on-source, others are off-source. 
        # This ensures that all on-source dithers are included in the "total_exposure_time", whereas 
        # off-source time is only reported in "all_dithers_time".
        total_exposure_time = r['information']['exposure_specification']['total_exposure_time']
        if 'n_on_source' in self.extracted:
            r['scalar']['total_exposure_time'] = total_exposure_time * self.extracted['n_on_source']
            r['scalar']['all_dithers_time'] = total_exposure_time * self.extracted['n_total']
        else:
            r['scalar']['total_exposure_time'] = total_exposure_time
            r['scalar']['all_dithers_time'] = total_exposure_time

        # Report on measurement, exposure time (time for a single exposure) and saturation time. 
        # The r['scalar']['exposure_time'] value reported back to the client is not used at this point.
        r['scalar']['exposure_time'] = r['information']['exposure_specification']['exposure_time']
        r['scalar']['measurement_time'] = r['information']['exposure_specification']['measurement_time']
        r['scalar']['saturation_time'] = r['information']['exposure_specification']['saturation_time']
        r['scalar']['total_integrations'] = r['information']['exposure_specification']['total_integrations']

        # Report the exposure duty cycle
        duty_cycle = r['information']['exposure_specification']['duty_cycle']
        r['scalar']['duty_cycle'] = duty_cycle

        # get the pixel CR rate in events/s and convert to events/ramp
        r['scalar']['cr_ramp_rate'] = self.noise.pix_cr_rate * r['information']['exposure_specification']['saturation_time']

        # put some instrument and strategy configuration info into the scalar report
        r['scalar']['filter'] = self.input['configuration']['instrument']['filter']
        r['scalar']['disperser'] = self.input['configuration']['instrument']['disperser']

        # target_xy is not found in MSASpecApPhot and may not be found in future strategies as well.
        # if it's not there, assume the target position is the center of the FOV, i.e. (0.0, 0.0).
        if 'target_xy' in self.input['strategy']:
            r['scalar']['x_offset'] = self.input['strategy']['target_xy'][0]
            r['scalar']['y_offset'] = self.input['strategy']['target_xy'][1]
        else:
            r['scalar']['x_offset'] = 0.0
            r['scalar']['y_offset'] = 0.0

        # not all strategies define an aperture size. MSASpecApPhot doesn't because the size is defined by an
        # MSA aperture. optimal extraction strategies likely will not use this, either, once implemented.
        if 'aperture_size' in self.input['strategy']:
            r['scalar']['aperture_size'] = self.input['strategy']['aperture_size']
        else:
            r['scalar']['aperture_size'] = "N/A"

        # return a copy of the input as well
        r['input'] = self.input

        # not yet fully implemented...
        r['warnings'] = self.warnings

        # count total number of saturated pixels in the scene, and warn if necessary
        n_partial_saturated = (self.saturation == 1).sum()
        n_full_saturated = (self.saturation == 2).sum()
        if n_partial_saturated > 0:
            # At some point, the word nonlinear should be changed in the api
            r['warnings']['nonlinear'] = "Partial saturation:\n There are %d saturated pixels at the end " % n_partial_saturated
            r['warnings']['nonlinear'] += "of a ramp. Partial ramps may still be used in some cases."
        if n_full_saturated > 0:
            r['warnings']['saturated'] = "Full saturation:\n There are %d saturated pixels at the end of the " % n_full_saturated
            r['warnings']['saturated'] += "first group. These pixels cannot be recovered."
        
        # Target acquisition mode checks             
        if self.input['configuration']['instrument']['mode'] == 'target_acq':

            # Target acquisitions allow some number of saturated pixels within the extraction aperture. We
            # count both the partial and full saturated pixels.
            number_saturated_pixels = r['1d']['n_full_saturated'][1][0]+r['1d']['n_partial_saturated'][1][0]
            if number_saturated_pixels > self.signal.current_instrument.max_saturated_pixels:
                msg = "<font color=red><b>TA MAY FAIL</b></font>: Number of fully saturated pixels {} in centroid" \
                      " box exceeds the maximum number of {} pixels allowed to ensure a successful " \
                      "target acquisition. Recommend to adjust Detector Setup and/or Instrument " \
                      "Setup accordingly.".format(
                    number_saturated_pixels,
                    self.signal.current_instrument.max_saturated_pixels
                )
                r['warnings'].update({'ta_max_saturated_pixels': msg})

            # Compare to the target acquisition SNR requirement. If target is fully saturated, there is no need to
            # check SNR (since it will be 0)
            if (r['scalar']['sn'] < self.signal.current_instrument.min_snr_threshold) & \
                    (not r['1d']['n_full_saturated'][1][0] > 0):
                msg = "<font color=red><b>TA MAY FAIL</b></font>: calculated SNR {:.2f} must " \
                      "exceed {} to ensure TA success.".format(
                    r['scalar']['sn'], self.signal.current_instrument.min_snr_threshold
                )
                r['warnings'].update({'ta_snr_threshold': msg})
        
        return r

    def as_fits(self):
        """
        This takes the output of self.as_dict() and generates pyfits PrimaryHDU objects
        from the image data.

        Returns
        -------
        output: dict
            dictionary containing pyfits HDUList versions of the image data in report, r.
            The keys match those used in the input.  r['2D']['Grid'] is not carried over,
            though, because the information it contains is incorporated in the FITS headers
            where required.
        """
        r = self.as_dict()

        # add WCS information from the signal that describes the detector plane coordinates.
        # this can be either spatial vs spatial or spatial vs wavelength.
        detector_header = self.pix_grid.wcs_info()

        output = copy.deepcopy(r)

        # handle the sub-reports
        output['sub_reports'] = []
        for report in self.sub_reports:
            output['sub_reports'].append(report.as_fits())

        # first tackle the 3D model cubes, 1 per aperture slice...
        for k in ['flux', 'flux_plus_background']:
            for i in range(len(r['3d'][k])):
                # the cube data coming out of pandeia.engine is ordered x,y,wavelength which is
                # backwards from the numpy convention of ordering axes from slowest to fastest.
                # it is set up this way because we need to broadcast 1D vectors (e.g. througput
                # vs. wavelength) onto the 3D cubes. for this to work, wavelength must be the 3rd
                # axis. so transpose the cube and then flip the Y axis to get the expected
                # orientation in the FITS data as viewed in DS9 or the like.
                o = fits.PrimaryHDU(r['3d'][k][i].transpose()[:, ::-1, :])
                tbhdu, header = self.signal.cube_wcs_info()
                o.header.update(header)
                output['3d'][k][i] = fits.HDUList([o, tbhdu])

        if r['input']['strategy']['method'] in ['ifuapphot', 'ifunodinscene', 'ifunodoffscene']:
            header = self.pix_grid.wcs_info()
            t = self.signal.spectral_detector_transform()
            header['ctype3'] = 'Wavelength'
            header['crpix3'] = 1
            header['crval3'] = t['wave_det_min'] - 0.5 * t['wave_det_step']
            header['cdelt3'] = t['wave_det_step']
            header['cunit3'] = 'um'
            header['cname3'] = 'Wavelength'
            for k in ['', '_signal', '_noise', '_snr', '_saturation']:
                key = 'reconstructed' + k
                # the reconstructed cubes are in the proper z, y, x order...
                o = fits.PrimaryHDU(r['3d'][key][::-1, ::-1, :])
                o.header.update(header)
                output['3d'][key] = o

        # now the 2D data
        for k in r['2d']:
            # flip the Y axis to get the FITS data to look the
            # same in DS9 as in matplotlib.
            o = fits.PrimaryHDU(r['2d'][k][::-1, :])
            o.header.update(detector_header)
            output['2d'][k] = o

        # now the 1D data as binary FITS tables
        for k in list(self.curves.keys()):
            tbhdu = fits.BinTableHDU.from_columns([
                fits.Column(name='WAVELENGTH',
                            unit='um',
                            format="1D",
                            array=self.curves[k][0]),
                fits.Column(name=k,
                            format="1D",
                            array=self.curves[k][1])
            ])
            tbhdu.name = k
            output['1d'][k] = fits.HDUList([tbhdu])
        for k in ['wave_calc', 'wave_pix']:
            tbhdu = fits.BinTableHDU.from_columns([
                fits.Column(name='WAVELENGTH',
                            unit='um',
                            format="1D",
                            array=output['1d'][k])
            ])
            tbhdu.name = k
            output['1d'][k] = fits.HDUList([tbhdu])

        return output





import contextlib
import os
import shutil
import stat

try:
    reload = reload
except NameError:
    from imp import reload

from configparser import ConfigParser
from distutils.ccompiler import new_compiler
from distutils.msvccompiler import MSVCCompiler
from distutils.sysconfig import customize_compiler


@contextlib.contextmanager
def open_config(filename):
    cfg = ConfigParser()
    cfg.read(filename)
    yield cfg
    with open(filename, 'w') as fp:
        cfg.write(fp)


def get_compiler_command():
    """
    Returns the name of the executable used by the default compiler on the
    system used by distutils to build C extensions.
    """

    compiler = new_compiler()
    customize_compiler(compiler)
    if isinstance(compiler, MSVCCompiler):
        compiler.initialize()
        # Normally the compiler path will be quoted as it contains spaces
        return '"%s"' % compiler.cc
    return compiler.compiler[0]


def rmtree(path):
    """
    shutil.rmtree() with error handler for 'access denied' from trying to
    delete read-only files.
    """

    def onerror(func, path, exc_info):
        if not os.access(path, os.W_OK):
            os.chmod(path, stat.S_IWUSR)
            func(path)
        else:
            raise

    return shutil.rmtree(path, onerror=onerror)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, absolute_import

import os
import copy

import astropy.io.fits as fits
import numpy as np

from .custom_exceptions import DataError


class PSFLibrary:

    """
    Class for encapsulating a hard PSF library, typically one generated
    by WebbPSF or its derivatives.

    Parameters
    ----------
    path : string, optional
        the path to the PSF library root (any .fits files in the
        path root will be assumed to be a valid PSF file in the WebbPSF format.
        If it is none, the PSF library from the package distribution will be used.
    aperture : string, optional
        restrict psfs to a specific mode with the name matching the keyword string.
        Strictly, the reader will look for psf file names with a substring matching the aperture
        string. 'all' reads all psfs in the path. This is the default. The aperture string in the filename
        should match the .fits keyword APERTURE.

    Methods
    -------
    read_library
    get_values
    get_good_psfs
    get_psf

    """

    def __init__(self, path=None, aperture='all'):
        if path is None:
            path = os.path.join(os.path.dirname(__file__), 'refdata', 'psfs')
        self.read_library(path, aperture)

    def read_library(self, path, aperture, wave_unit='m'):
        """
        Read in a library of PSFs stored in FITS files. The filenames contain information about each PSF. The fields are
        separated by _'s with the resulting list entries corresponding to:
            0 - instrument name
            1 - apertures for which PSF is used. this is a - separated string that is split() into a list of apertures
            2 - wavelength at which PSF is calculated
            3, 4 (optional) - r,theta polar coordinates in FOV at which at which PSF is calculated

        Parameters
        ----------
        path: string
            Path containing the PSF FITS images

        wave_unit: string
            Wavelength unit used by PSF library (default: m)
        """
        files = os.listdir(path)

        psf_files = []
        for file in files:
            filename, ext = os.path.splitext(file)
            filename_split = filename.split('_')
            filename_instrument = filename_split[0]
            filename_apertures = filename_split[1].split('-')
            filename_wave = filename_split[2]
            # If the psf is on-axis, the offset polar coordinates may not exist
            try:
                filename_radius = filename_split[3]
                filename_theta = filename_split[4]
            except:
                filename_radius = 0.
                filename_theta = 0.

            if aperture is 'all':
                if ext == '.fits':
                    psf_files.append(file)
            elif ext == '.fits' and aperture in filename_apertures:
                    psf_files.append(file)
            # If the aperture has a string with an underscore in it, try to use only the first part.
            elif ext == '.fits' and (aperture.split('_'))[0] in filename_apertures:
                    psf_files.append(file)
            else:
                pass

        self._psfs = []

        for psf_file in psf_files:
            hdulist = fits.open(path + '/' + psf_file, memmap=False)

            psf_int = copy.deepcopy(hdulist[0].data)

            ins = hdulist[0].header['INSTRUME'].lower()
            nwaves = hdulist[0].header['NWAVES']
            if nwaves != 1:
                raise ValueError("Input PSF must be monochromatic (nwaves = #r)" % nwaves)

            wave_scl = self._get_unit_scl(wave_unit)
            wave = hdulist[0].header['WAVE0'] * wave_scl

            pix_scl = hdulist[0].header['PIXELSCL']
            diff_limit = hdulist[0].header['DIFFLMT']
            aperture_name = hdulist[0].header['APERTURE'].lower()
            # offset_r/offset_t is the offset from the optical center in webbpsf.  however, for NIRCam masklwb
            # and maskswb, this includes the offset from the optical center to the optimal position along the bar.
            # what we need for source association is instead the offset from that optimal position. this is stored in the
            # optoff_r/optoff_t keywords which need to be provided in these cases. aperture_name here will have these
            # aperture and filter concatenated so do a string compare to look for aperture.
            if 'masklwb' in aperture_name or 'maskswb' in aperture_name:
                try:
                    radius = hdulist[0].header['OPTOFF_R']
                    theta = hdulist[0].header['OPTOFF_T']
                except KeyError as e:
                    msg = "PSFs for bar-shaped masks require offsets from optimal bar position. (%s)" % e
                    raise DataError(value=msg)
            else:
                radius = hdulist[0].header['OFFSET_R']
                theta = hdulist[0].header['OFFSET_T']

            upsamp = hdulist[0].header['DET_SAMP']

            psf = {
                'int': psf_int,
                'wave': wave,
                'pix_scl': pix_scl,
                'diff_limit': diff_limit,
                'upsamp': upsamp,
                'instrument': ins,
                'aperture_name': aperture_name,
                'source_offset': (radius, theta)
            }

            self._psfs.append(psf)
            hdulist.close()

    def get_values(self, key, instrument, aperture_name, source_offset=(0, 0)):
        """
        Returns the available values of a given key for a given instrument and aperture name.

        Parameters
        ----------
        key: string
            Desired key. One of 'int', 'wave', 'pix_scl', 'diff_limit', 'upsamp', 'instrument',
            'aperture_name', 'source_offset_r', or 'source_offset_theta'

        Returns
        -------
        result: tuple of format (list, list)
            Lists of index id's and values
        """
        values = [psf[key] for psf in self._psfs
                  if (instrument == psf['instrument'] and
                      aperture_name in psf['aperture_name']) and
                      source_offset == psf['source_offset']]
        ids = [i for i, psf in enumerate(self._psfs)
               if (instrument == psf['instrument'] and
                   aperture_name in psf['aperture_name']) and
                   source_offset == psf['source_offset']]
        result = ids, values
        return result

    def get_psf(self, wave, instrument, aperture_name, source_offset=(0, 0)):
        """
        Get PSF given wavelength, instrument, instrument aperture, and optionally source offset polar coordinates.
        Values are interpolated linearly between the two PSF library entries that bracket the given wavelength.

        Parameters
        ----------
        wave: float
            Desired wavelength for the PSF
        instrument: str
            Instrument name
        aperture_name: str
            Name of the instrument aperture

        Returns
        -------
        psf: dict
            Dict containing the PSF and associated information
        """
        wids, psf_waves = self.get_values('wave', instrument, aperture_name, source_offset=source_offset)
        nids, nearest_waves = self._find_two_nearest(psf_waves, wave)

        # if wave is not in psf_waves then the two closest will be returned
        if len(nids) == 2:
            ids = [wids[nids[0]], wids[nids[1]]]

            pix_scl0 = self._psfs[ids[0]]['pix_scl']
            pix_scl1 = self._psfs[ids[1]]['pix_scl']
            if pix_scl0 != pix_scl1:
                raise ValueError("Pixel scales in the library must be the same for a single instrument aperture.")

            psf_int0 = self._psfs[ids[0]]['int']
            psf_int1 = self._psfs[ids[1]]['int']
            wave0 = self._psfs[ids[0]]['wave']
            wave1 = self._psfs[ids[1]]['wave']
            diff_limit0 = self._psfs[ids[0]]['diff_limit']
            diff_limit1 = self._psfs[ids[1]]['diff_limit']

            upsamp0 = self._psfs[ids[0]]['upsamp']
            upsamp1 = self._psfs[ids[1]]['upsamp']
            if upsamp0 != upsamp1:
                raise ValueError("Upsampling factors in the library must be the same for a single instrument aperture.")

            psf_int = psf_int0 + (psf_int1 - psf_int0) * (wave - wave0) / (wave1 - wave0)
            diff_limit = diff_limit0 + (diff_limit1 - diff_limit0) * (wave - wave0) / (wave1 - wave0)

        # if wave is one of the values in psf_waves then only it will be returned and
        # no interpolation is necessary
        else:
            ids = nids

            psf_int = self._psfs[ids[0]]['int']
            diff_limit =  self._psfs[ids[0]]['diff_limit']
            pix_scl0 = self._psfs[ids[0]]['pix_scl']
            upsamp0 = self._psfs[ids[0]]['upsamp']

        psf = {
            'int': psf_int,
            'wave': wave,
            'pix_scl': pix_scl0,
            'diff_limit': diff_limit,
            'upsamp': upsamp0,
            'instrument': self._psfs[ids[0]]['instrument'],
            'aperture_name': self._psfs[ids[0]]['aperture_name'],
            'source_offset': self._psfs[ids[0]]['source_offset']
        }
        return psf

    def get_pix_scale(self, instrument, aperture_name):
        """
        Get PSF pixel scale for given instrument/aperture

        Parameters
        ----------
        instrument: str
            Instrument name
        aperture_name: str
            Name of instrument aperture

        Returns
        -------
        pix_scl: float
            Pixel scale of the PSF in arcsec/pixel
        """
        wids, psf_waves = self.get_values('wave', instrument, aperture_name)
        pix_scl = self._psfs[wids[0]]['pix_scl']
        return pix_scl

    def get_shape(self, instrument, aperture_name):
        """
        Get PSF shape for given instrument/aperture

        Parameters
        ----------
        instrument: str
            Instrument name
        aperture_name: str
            Name of instrument aperture

        Returns
        -------
        sh: tuple (int, int)
            Shape of the PSF kernel image
        """
        wids, psf_waves = self.get_values('wave', instrument, aperture_name)
        sh = self._psfs[wids[0]]['int'].shape
        return sh

    def get_upsamp(self, instrument, aperture_name):
        """
        Get PSF upsampling for given instrument/aperture

        Parameters
        ----------
        instrument: str
            Instrument name
        aperture_name: str
            Name of instrument aperture

        Returns
        -------
        upsamp: int
            PSF upsampling factor
        """
        wids, psf_waves = self.get_values('wave', instrument, aperture_name)
        upsamp = self._psfs[wids[0]]['upsamp']
        return upsamp

    def get_offsets(self, instrument, aperture_name):
        """
        Get available PSF offset positions. PSF libraries supporting position-dependent
        PSFs will have multiple. Libraries that do not support position-dependent PSFs will
        only have one position.

        Parameters
        ----------
        list: list-like
            List of values to search
        value: float
            Input value to be bracketed

        Returns
        -------
        psf_positions: list of tuples (source_offset_r, source_offset_theta)
            Polar coordinates of PSF offsets.

        """
        offsets = [psf['source_offset'] for psf in self._psfs if instrument ==
                   psf['instrument'] and aperture_name in psf['aperture_name']]
        unique_offsets = list(set(offsets))
        return unique_offsets

    def associate_offset_to_source(self, sources, instrument, aperture_name):

        psf_offsets = self.get_offsets(instrument, aperture_name)
        psf_associations = []
        for source in sources:
            source_offset_radius = np.sqrt(source.position['x_offset']**2. + source.position['y_offset']**2.)
            # Currently, we only associate radius, not angle.
            distances = [np.abs(source_offset_radius-psf_offset[0]) for psf_offset in psf_offsets]
            closest_index = np.argmin(distances)
            psf_associations.append(psf_offsets[closest_index])

        return psf_associations

    def _find_two_nearest(self, values, value):
        """
        Find the subscripts of the two neighboring values in list relative to
        an input value.

        Parameters
        ----------
        values: list-like
            List of values to search
        value: float
            Input value to be bracketed

        Returns
        -------
        vals: tuple (list, np.ndarray)
            Bracketing indices in both list and numpy array format
        """
        nplist = np.array(values)

        # ensure that value is not outside the range of list
        if value < nplist.min() or value > nplist.max():
            raise ValueError("Value must be within range of list.")

        diff = nplist - value
        diff_sorted = np.sort(diff)

        smaller = np.max(diff_sorted[diff_sorted<=0])
        larger = np.min(diff_sorted[diff_sorted>=0])

        ids = [i for i, v in enumerate(diff) if diff[i] == smaller or diff[i] == larger]

        vals = ids, nplist[ids]

        return vals

    def _get_unit_scl(self, unit):
        """
        Get unit scale factor for given unit string: m, micron, nm, or Angstrom.
        This should get replaced with real astropy.units support eventually.

        Parameters
        ----------
        unit: str
            Unit string

        Returns
        -------
        scale: float
            Scale factor corresponding to unit string
        """
        scales = {'m': 1.0e6, 'micron': 1.0, 'nm': 1.0e-3, 'Angstrom': 1.0e-4}
        try:
            scale = scales[unit]
            return scale
        except KeyError:
            raise KeyError('Unknown wavelength unit')
